# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2015-2023 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""APT key management helpers."""

# pyright: reportMissingTypeStubs=false

import logging
import pathlib
import subprocess
import tempfile
from contextlib import contextmanager
from typing import Iterable, Iterator, List, Optional

from . import apt_ppa, errors, package_repository

logger = logging.getLogger(__name__)

DEFAULT_APT_KEYSERVER = "keyserver.ubuntu.com"

# Directory for apt keyrings as recommended by Debian for third-party keyrings.
KEYRINGS_PATH = pathlib.Path("/etc/apt/keyrings")

# GnuPG command line options that we always want to use.
_GPG_PREFIX = ["gpg", "--batch", "--no-default-keyring", "--with-colons"]


def _call_gpg(
    *parameters: str,
    keyring: Optional[pathlib.Path] = None,
    base_parameters: Iterable[str] = _GPG_PREFIX,
    stdin: Optional[bytes] = None,
) -> bytes:
    if keyring:
        command = [*base_parameters, "--keyring", f"gnupg-ring:{keyring}", *parameters]
    else:
        command = [*base_parameters, *parameters]
    logger.debug(f"Executing command: {command}")
    env = {"LANG": "C.UTF-8"}
    process = subprocess.run(
        command,
        input=stdin,
        capture_output=True,
        check=True,
        env=env,
    )
    return process.stdout


def get_keyring_path(
    key_id: str,
    *,
    is_ascii: bool = False,
    base_path: pathlib.Path = KEYRINGS_PATH,
    prefix: str = "craft-",
) -> pathlib.Path:
    """Get a Path object where we would expect to find a key.

    :param key_id: The key ID for the keyring file.
    :param base_path: The directory for the key.
    :param prefix: The prefix fer the keyfile
    :param is_ascii: Whether the file is ASCII-armored (.asc suffix)

    :returns: A Path object matching the expected filename
    """
    file_base = prefix + key_id[-8:].upper()
    return base_path.joinpath(file_base).with_suffix(".asc" if is_ascii else ".gpg")


def _configure_keyring_file(keyring_file: pathlib.Path) -> None:
    """Configure a newly created keyring file."""
    # Change the permissions on the file so that APT itself can read it later
    keyring_file.chmod(0o644)
    # Also remove the backup file, if gpg created it
    backup = keyring_file.with_suffix(keyring_file.suffix + "~")
    backup.unlink(missing_ok=True)


class AptKeyManager:
    """Manage APT repository keys."""

    def __init__(
        self,
        *,
        keyrings_path: Optional[pathlib.Path] = None,
        key_assets: pathlib.Path,
    ) -> None:
        self._keyrings_path = keyrings_path or KEYRINGS_PATH
        self._key_assets = key_assets

    def find_asset_with_key_id(self, *, key_id: str) -> Optional[pathlib.Path]:
        """Find snap key asset matching key_id.

        The key asset much be named with the last 8 characters of the key
        identifier, in upper case.

        :param key_id: Key ID to search for.

        :returns: Path of key asset if match found, otherwise None.
        """
        key_path = get_keyring_path(
            key_id, is_ascii=True, prefix="", base_path=self._key_assets
        )

        if key_path.exists():
            return key_path

        return None

    @classmethod
    def keyrings_path_for_root(
        cls, root: Optional[pathlib.Path] = None
    ) -> pathlib.Path:
        """Get the location for Apt keyrings with ``root`` as the system root.

        :param root: The optional system root to consider, or None to assume the standard
          system root "/".
        """
        if root is None:
            return KEYRINGS_PATH
        return root / "etc/apt/keyrings"

    @classmethod
    def get_key_fingerprints(cls, *, key: str) -> List[str]:
        """List fingerprints found in specified key.

        :param key: Key data (string) to parse.

        :returns: List of key fingerprints/IDs.
        """
        with _temporary_home_dir() as tmpdir:
            response = _call_gpg(
                "--homedir",
                str(tmpdir),
                "--import-options",
                "show-only",
                "--import",
                stdin=key.encode(),
            ).splitlines()
        fingerprints: List[str] = []
        for line in response:
            if line.startswith(b"fpr:"):
                fingerprints.append(line[4:].decode().strip(":"))
        return fingerprints

    def is_key_installed(self, *, key_id: str) -> bool:
        """Check if specified key_id is installed.

        :param key_id: Key ID to check for. The key will be looked for in the
          AptKeyManager's configured keyrings path.

        :returns: True if key is installed.
        """
        keyring_file = get_keyring_path(key_id, base_path=self._keyrings_path)
        # Check if the keyring file exists first, otherwise the gpg check itself
        # creates it.
        if not keyring_file.is_file():
            logger.debug(f"Keyring file not found: {keyring_file}")
            return False

        # Ensure the keyring file contains the correct key
        try:
            logger.debug("Listing keys in keyring...")
            _call_gpg("--list-keys", key_id, keyring=keyring_file)
        except subprocess.CalledProcessError as error:
            logger.warning(f"gpg error: {error.output.decode()}")
            return False
        else:
            return True

    def install_key(self, *, key: str) -> None:
        """Install given key.

        :param key: Key to install.

        :raises: AptGPGKeyInstallError if unable to install key.
        """
        logger.debug(f"Importing key {key}")
        fingerprints = self.get_key_fingerprints(key=key)
        if not fingerprints:
            raise errors.AptGPGKeyInstallError("Invalid GPG key", key=key)
        if len(fingerprints) != 1:
            raise errors.AptGPGKeyInstallError(
                "Key must be a single key, not multiple.", key=key
            )

        self._create_keyrings_path()
        keyring_path = get_keyring_path(fingerprints[0], base_path=self._keyrings_path)
        try:
            _call_gpg("--import", "-", keyring=keyring_path, stdin=key.encode())
        except subprocess.CalledProcessError as error:
            raise errors.AptGPGKeyInstallError(error.output.decode(), key=key)

        _configure_keyring_file(keyring_path)
        logger.debug(f"Installed apt repository key:\n{key}")

    def install_key_from_keyserver(
        self, *, key_id: str, key_server: str = DEFAULT_APT_KEYSERVER
    ) -> None:
        """Install key from specified key server.

        :param key_id: Key ID to install.
        :param key_server: Key server to query.

        :raises: AptGPGKeyInstallError if unable to install key.
        """
        self._create_keyrings_path()
        keyring_path = get_keyring_path(key_id, base_path=self._keyrings_path)
        try:
            with _temporary_home_dir() as tmpdir:
                # We use a tmpdir because gpg needs a "homedir" to place temporary
                # files into during the download process.
                _call_gpg(
                    "--homedir",
                    str(tmpdir),
                    "--keyserver",
                    key_server,
                    "--recv-keys",
                    key_id,
                    keyring=keyring_path,
                )
            _configure_keyring_file(keyring_path)
        except subprocess.CalledProcessError as error:
            raise errors.AptGPGKeyInstallError(
                error.output.decode(), key_id=key_id, key_server=key_server
            )

    def install_package_repository_key(
        self, *, package_repo: package_repository.PackageRepository
    ) -> bool:
        """Install required key for specified package repository.

        For both PPA and other Apt package repositories:
        1) If key is already installed, return False.
        2) Install key from local asset, if available.
        3) Install key from key server, if available. An unspecified
           keyserver will default to using keyserver.ubuntu.com.

        :param package_repo: Apt PackageRepository configuration.

        :returns: True if key configuration was changed. False if
            key already installed.

        :raises: AptGPGKeyInstallError if unable to install key.
        """
        key_server = DEFAULT_APT_KEYSERVER
        if isinstance(package_repo, package_repository.PackageRepositoryAptPPA):
            key_id = apt_ppa.get_launchpad_ppa_key_id(ppa=package_repo.ppa)
        elif isinstance(package_repo, package_repository.PackageRepositoryAptUCA):
            key_id = package_repository.UCA_KEY_ID
        elif isinstance(package_repo, package_repository.PackageRepositoryApt):
            key_id = package_repo.key_id
            if package_repo.key_server:
                key_server = package_repo.key_server
        else:
            raise RuntimeError(f"unhandled package repo type: {package_repo!r}")

        # Already installed, nothing to do.
        if self.is_key_installed(key_id=key_id):
            return False

        # If the keyring exists but does not contain the key, remove it and
        # install a fresh one.
        keyring_path = get_keyring_path(key_id, base_path=self._keyrings_path)
        if keyring_path.parent.is_dir():
            keyring_path.unlink(missing_ok=True)

        key_path = self.find_asset_with_key_id(key_id=key_id)
        if key_path is not None:
            self.install_key(key=key_path.read_text())
        else:
            self.install_key_from_keyserver(key_id=key_id, key_server=key_server)

        return True

    def _create_keyrings_path(self) -> None:
        """Create the directory that will contain the keys, if necessary."""
        if not self._keyrings_path.exists():
            logger.debug(
                f"Keyrings location {self._keyrings_path} doesn't exist; Attempting to create it."
            )
            self._keyrings_path.mkdir(mode=0o755, parents=False)


@contextmanager
def _temporary_home_dir() -> Iterator[pathlib.Path]:
    """Yield a temporary directory with proper permissions for gpg."""
    with tempfile.TemporaryDirectory() as tmpdir_str:
        tmpdir = pathlib.Path(tmpdir_str)
        tmpdir.chmod(0o700)
        yield tmpdir



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "craft-archives"
copyright = "2023, Canonical"
author = "Canonical"

# region General configuration
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx_design",
    "sphinx_copybutton",
    "sphinx-pydantic",
    "sphinx_toolbox",
    "sphinx_toolbox.more_autodoc",
    "sphinx.ext.autodoc",  # Must be loaded after more_autodoc
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

show_authors = False

# endregion
# region Options for HTML output
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_static_path = ["_static"]
html_css_files = [
    "css/custom.css",
]

# endregion
# region Options for extensions
# Intersphinx extension
# https://www.sphinx-doc.org/en/master/usage/extensions/intersphinx.html#configuration

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
}

# Type hints configuration
set_type_checking_flag = True
typehints_fully_qualified = False
always_document_param_types = True

# Github config
github_username = "canonical"
github_repository = "craft-archives"

# endregion



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2019-2023 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import pydantic
import pytest
from craft_archives.repo import errors
from craft_archives.repo.package_repository import (
    PackageRepository,
    PackageRepositoryApt,
    PackageRepositoryAptPPA,
    PackageRepositoryAptUCA,
)

# pyright: reportGeneralTypeIssues=false

# region Test data and fixtures
BASIC_PPA_MARSHALLED = {"type": "apt", "ppa": "test/foo", "priority": 123}
BASIC_UCA_MARSHALLED = {
    "type": "apt",
    "cloud": "antelope",
    "pocket": "updates",
    "priority": 123,
}
BASIC_APT_MARSHALLED = {
    "architectures": ["amd64", "i386"],
    "components": ["main", "multiverse"],
    "formats": ["deb", "deb-src"],
    "key-id": "A" * 40,
    "key-server": "keyserver.ubuntu.com",
    "suites": ["xenial", "xenial-updates"],
    "type": "apt",
    "url": "http://archive.ubuntu.com/ubuntu",
    "priority": 123,
}


@pytest.fixture
def apt_repository():
    yield PackageRepositoryApt(
        type="apt",
        architectures=["amd64", "i386"],
        components=["main", "multiverse"],
        formats=["deb", "deb-src"],
        key_id="A" * 40,
        key_server="keyserver.ubuntu.com",
        # name="test-name",
        suites=["xenial", "xenial-updates"],
        url="http://archive.ubuntu.com/ubuntu",
        priority=123,
    )


# endregion
# region PackageRepositoryApt
def create_apt(**kwargs) -> PackageRepositoryApt:
    return PackageRepositoryApt(type="apt", **kwargs)


def test_apt_name():
    repo = create_apt(
        key_id="A" * 40,
        url="http://archive.ubuntu.com/ubuntu",
    )
    assert repo.name == "http_archive_ubuntu_com_ubuntu"


@pytest.mark.parametrize(
    "priority", ["always", "prefer", "defer", 1000, 990, 500, 100, -1, None]
)
@pytest.mark.parametrize(
    "repo",
    [
        {
            "type": "apt",
            "url": "https://some/url",
            "key-id": "BCDEF12345" * 4,
            "path": "my/path",
        },
        {
            "type": "apt",
            "url": "https://some/url",
            "key-id": "BCDEF12345" * 4,
            "formats": ["deb"],
            "components": ["some", "components"],
            "key-server": "my-key-server",
            "suites": ["some", "suites"],
        },
    ],
)
def test_apt_valid(repo, priority):
    if priority is not None:
        repo["priority"] = priority
    apt_deb = PackageRepositoryApt.unmarshal(repo)
    assert apt_deb.type == "apt"
    assert apt_deb.url == "https://some/url"
    assert apt_deb.key_id == "BCDEF12345" * 4
    assert apt_deb.formats == (["deb"] if "formats" in repo else None)
    assert apt_deb.components == (
        ["some", "components"] if "components" in repo else None
    )
    assert apt_deb.key_server == ("my-key-server" if "key-server" in repo else None)
    assert apt_deb.path == ("my/path" if "path" in repo else None)
    assert apt_deb.suites == (["some", "suites"] if "suites" in repo else None)


@pytest.mark.parametrize(
    "arch", ["amd64", "armhf", "arm64", "i386", "ppc64el", "riscv", "s390x"]
)
def test_apt_valid_architectures(arch):
    package_repo = create_apt(key_id="A" * 40, url="http://test", architectures=[arch])

    assert package_repo.architectures == [arch]


def test_apt_invalid_url():
    with pytest.raises(
        pydantic.ValidationError, match="Invalid URL; URLs must be non-empty strings"
    ):
        create_apt(
            key_id="A" * 40,
            url="",
        )


def test_apt_invalid_path():
    with pytest.raises(
        pydantic.ValidationError, match="Invalid path; Paths must be non-empty strings."
    ):
        create_apt(
            key_id="A" * 40,
            path="",
            url="http://archive.ubuntu.com/ubuntu",
        )


def test_apt_invalid_path_with_suites():
    with pytest.raises(pydantic.ValidationError) as raised:
        create_apt(
            key_id="A" * 40,
            path="/",
            suites=["xenial", "xenial-updates"],
            url="http://archive.ubuntu.com/ubuntu",
        )

    expected_message = (
        "Invalid package repository for 'http://archive.ubuntu.com/ubuntu': "
        "suites ['xenial', 'xenial-updates'] cannot be combined with path '/'"
    )

    err = raised.value
    assert expected_message in str(err)


def test_apt_invalid_path_with_components():
    with pytest.raises(pydantic.ValidationError) as raised:
        create_apt(
            key_id="A" * 40,
            path="/",
            components=["main"],
            url="http://archive.ubuntu.com/ubuntu",
        )

    expected_message = (
        "Invalid package repository for 'http://archive.ubuntu.com/ubuntu': "
        "components ['main'] cannot be combined with path '/'."
    )

    err = raised.value
    assert expected_message in str(err)


def test_apt_invalid_missing_components():
    with pytest.raises(pydantic.ValidationError) as raised:
        create_apt(
            key_id="A" * 40,
            suites=["xenial", "xenial-updates"],
            url="http://archive.ubuntu.com/ubuntu",
        )

    expected_message = (
        "Invalid package repository for 'http://archive.ubuntu.com/ubuntu': "
        "components must be specified when using suites."
    )

    err = raised.value
    assert expected_message in str(err)


def test_apt_invalid_missing_suites():
    with pytest.raises(pydantic.ValidationError) as raised:
        create_apt(
            key_id="A" * 40,
            components=["main"],
            url="http://archive.ubuntu.com/ubuntu",
        )

    expected_message = (
        "Invalid package repository for 'http://archive.ubuntu.com/ubuntu': "
        "suites must be specified when using components."
    )

    err = raised.value
    assert expected_message in str(err)


def test_apt_invalid_suites_as_path():
    with pytest.raises(pydantic.ValidationError) as raised:
        create_apt(
            key_id="A" * 40,
            suites=["my-suite/"],
            url="http://archive.ubuntu.com/ubuntu",
        )

    expected_message = (
        "Invalid package repository for 'http://archive.ubuntu.com/ubuntu': "
        "invalid suite 'my-suite/'. Suites must not end with a '/'."
    )

    err = raised.value
    assert expected_message in str(err)


def test_apt_key_id_valid():
    key_id = "ABCDE12345" * 4
    repo = {
        "type": "apt",
        "url": "https://some/url",
        "key-id": key_id,
    }
    apt_deb = PackageRepositoryApt.unmarshal(repo)
    assert apt_deb.key_id == key_id


@pytest.mark.parametrize(
    "key_id",
    ("KEYID12345" * 4, "abcde12345" * 4),
)
def test_apt_key_id_invalid(key_id):
    repo = {
        "type": "apt",
        "url": "https://some/url",
        "key-id": key_id,
    }

    with pytest.raises(pydantic.ValidationError, match="string does not match regex"):
        PackageRepositoryApt.unmarshal(repo)


@pytest.mark.parametrize(
    "formats",
    [
        ["deb"],
        ["deb-src"],
        ["deb", "deb-src"],
        ["_invalid"],
    ],
)
def test_apt_formats(formats):
    repo = {
        "type": "apt",
        "url": "https://some/url",
        "key-id": "ABCDE12345" * 4,
        "formats": formats,
    }

    if formats != ["_invalid"]:
        apt_deb = PackageRepositoryApt.unmarshal(repo)
        assert apt_deb.formats == formats
    else:
        error = ".*unexpected value; permitted: 'deb', 'deb-src'"
        with pytest.raises(pydantic.ValidationError, match=error):
            PackageRepositoryApt.unmarshal(repo)


def test_apt_marshal(apt_repository):
    assert apt_repository.marshal() == BASIC_APT_MARSHALLED


def test_apt_unmarshal_invalid_extra_keys():
    test_dict = {
        "architectures": ["amd64", "i386"],
        "components": ["main", "multiverse"],
        "formats": ["deb", "deb-src"],
        "key-id": "A" * 40,
        "key-server": "keyserver.ubuntu.com",
        "name": "test-name",
        "suites": ["xenial", "xenial-updates"],
        "type": "apt",
        "url": "http://archive.ubuntu.com/ubuntu",
        "priority": 123,
        "foo": "bar",
        "foo2": "bar",
    }

    with pytest.raises(pydantic.ValidationError):
        PackageRepositoryApt.unmarshal(test_dict)


def test_apt_unmarshal_invalid_type():
    test_dict = {
        "architectures": ["amd64", "i386"],
        "components": ["main", "multiverse"],
        "formats": ["deb", "deb-src"],
        "key-id": "A" * 40,
        "key-server": "keyserver.ubuntu.com",
        # "name": "test-name",
        "suites": ["xenial", "xenial-updates"],
        "type": "aptx",
        "url": "http://archive.ubuntu.com/ubuntu",
        "priority": "always",
    }

    with pytest.raises(pydantic.ValidationError):
        PackageRepositoryApt.unmarshal(test_dict)


@pytest.mark.parametrize(
    "repository",
    [
        BASIC_APT_MARSHALLED,
        {
            "type": "apt",
            "key-id": "A" * 40,
            "url": "https://example.com",
        },
        {
            "type": "apt",
            "key-id": "A" * 40,
            "url": "https://example.com",
            "architectures": ["spookyarch"],
        },
        {
            "type": "apt",
            "key-id": "A" * 40,
            "url": "https://example.com",
            "components": ["main"],
            "suites": ["jammy"],
        },
        {
            "type": "apt",
            "key-id": "A" * 40,
            "url": "https://example.com",
            "path": "/dev/null",
        },
        {
            "type": "apt",
            "key-id": "A" * 40,
            "url": "https://example.com",
            "components": ["main"],
            "suites": ["jammy"],
            "priority": 1234,
        },
    ],
)
def test_apt_marshal_unmarshal_inverses(repository):
    assert PackageRepositoryApt.unmarshal(repository).marshal() == repository


def test_apt_invalid_priority():
    with pytest.raises(pydantic.ValidationError) as raised:
        create_apt(key_id="A" * 40, url="http://test", priority=0)

    expected_message = (
        "Invalid package repository for 'http://test': "
        "invalid priority: Priority cannot be zero."
    )

    err = raised.value
    assert expected_message in str(err)


@pytest.mark.parametrize(
    "priority_str,priority_int",
    [
        ("always", 1000),
        ("prefer", 990),
        ("defer", 100),
    ],
)
def test_priority_correctly_converted(priority_str, priority_int):
    repo_marshalled = BASIC_APT_MARSHALLED.copy()
    repo_marshalled["priority"] = priority_str
    repo = PackageRepositoryApt.unmarshal(repo_marshalled)

    assert repo.priority == priority_int


@pytest.mark.parametrize(
    "url,pin",
    [
        ("https://example.com/repo", 'origin "example.com"'),
        ("http://archive.debian.org/debian/stable/blah", 'origin "archive.debian.org"'),
        (
            "https://ppa.launchpadcontent.net/deadsnakes/ppa/ubuntu/",
            'origin "ppa.launchpadcontent.net"',
        ),
    ],
)
def test_pin_value(url, pin):
    repo_marshalled = BASIC_APT_MARSHALLED.copy()
    repo_marshalled["url"] = url
    repo = PackageRepositoryApt.unmarshal(repo_marshalled)

    assert repo.pin == pin


# endregion
# region PackageRepositoryAptPPA
def create_ppa(**kwargs) -> PackageRepositoryAptPPA:
    return PackageRepositoryAptPPA(type="apt", **kwargs)


def test_ppa_marshal():
    repo = create_ppa(ppa="test/ppa", priority=123)

    assert repo.marshal() == {"type": "apt", "ppa": "test/ppa", "priority": 123}


def test_ppa_invalid_ppa():
    with pytest.raises(pydantic.ValidationError) as raised:
        create_ppa(ppa="")

    expected_message = "Invalid PPA: PPAs must be non-empty strings."
    err = raised.value
    assert expected_message in str(err)


@pytest.mark.parametrize(
    "priority_str,priority_int",
    [
        ("always", 1000),
        ("prefer", 990),
        ("defer", 100),
    ],
)
def test_ppa_priority_correctly_converted(priority_str, priority_int):
    repo_marshalled = BASIC_PPA_MARSHALLED.copy()
    repo_marshalled["priority"] = priority_str
    repo = PackageRepositoryAptPPA.unmarshal(repo_marshalled)

    assert repo.priority == priority_int


@pytest.mark.parametrize(
    "ppa,pin",
    [
        ("ppa/ppa", "release o=LP-PPA-ppa-ppa"),
        ("deadsnakes/nightly", "release o=LP-PPA-deadsnakes-nightly"),
    ],
)
def test_ppa_pin_value(ppa, pin):
    repo = PackageRepositoryAptPPA.unmarshal(
        {
            "type": "apt",
            "ppa": ppa,
        }
    )

    assert repo.pin == pin


# endregion
# region PackageRepository
@pytest.mark.parametrize("data", [None, "some_string"])
def test_unmarshal_validation_error(data):
    with pytest.raises(errors.PackageRepositoryValidationError) as raised:
        PackageRepository.unmarshal(data)

    assert (
        raised.value.details == "Package repository must be a valid dictionary object."
    )


@pytest.mark.parametrize(
    "repositories",
    [
        [],
        pytest.param([BASIC_PPA_MARSHALLED], id="ppa"),
        pytest.param([BASIC_APT_MARSHALLED], id="apt"),
        pytest.param([BASIC_APT_MARSHALLED, BASIC_PPA_MARSHALLED], id="ppa_and_apt"),
    ],
)
def test_marshal_unmarshal_inverses(repositories):
    objects = PackageRepository.unmarshal_package_repositories(repositories)
    marshalled = [repo.marshal() for repo in objects]

    assert marshalled == repositories


def test_unmarshal_package_repositories_list_none():
    assert PackageRepository.unmarshal_package_repositories(None) == []


def test_unmarshal_package_repositories_invalid_data():
    with pytest.raises(errors.PackageRepositoryValidationError) as raised:
        PackageRepository.unmarshal_package_repositories(
            "not-a-list"
        )  # pyright: ignore[reportGeneralTypeIssues]

    err = raised.value
    assert str(err) == (
        "Invalid package repository for 'not-a-list': invalid list object."
    )
    assert err.details == "Package repositories must be a list of objects."
    assert err.resolution == (
        "Verify 'package-repositories' configuration and ensure that "
        "the correct syntax is used."
    )


# endregion
# region PackageRepositoryAptCloud
def create_uca(**kwargs) -> PackageRepositoryAptUCA:
    return PackageRepositoryAptUCA(type="apt", **kwargs)


def test_uca_marshal():
    repo = create_uca(cloud="antelope", priority=123)

    assert repo.marshal() == {
        "type": "apt",
        "cloud": "antelope",
        "pocket": "updates",
        "priority": 123,
    }


def test_uca_invalid_cloud():
    with pytest.raises(pydantic.ValidationError) as raised:
        create_uca(cloud="")

    expected_message = "clouds must be non-empty strings."

    err = raised.value
    assert expected_message in str(err)


@pytest.mark.parametrize(
    "priority_str,priority_int",
    [
        ("always", 1000),
        ("prefer", 990),
        ("defer", 100),
    ],
)
def test_uca_priority_correctly_converted(priority_str, priority_int):
    repo_marshalled = BASIC_UCA_MARSHALLED.copy()
    repo_marshalled["priority"] = priority_str
    repo = PackageRepositoryAptUCA.unmarshal(repo_marshalled)

    assert repo.priority == priority_int


@pytest.mark.parametrize(
    "cloud,pin",
    [
        ("antelope", 'origin "ubuntu-cloud.archive.canonical.com"'),
        ("zed", 'origin "ubuntu-cloud.archive.canonical.com"'),
    ],
)
def test_uca_pin_value(cloud, pin):
    repo = PackageRepositoryAptUCA.unmarshal(
        {
            "type": "apt",
            "cloud": cloud,
        }
    )

    assert repo.pin == pin


# endregion



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2023 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Tests for apt_preferencs_manager"""
import shutil
from pathlib import Path
from textwrap import dedent

import pytest
from craft_archives.repo.apt_preferences_manager import (
    _DEFAULT_PREFERENCES_FILE,
    AptPreferencesManager,
    Preference,
)
from craft_archives.repo.errors import AptPreferencesError

VALID_PRIORITY_STRINGS = ("always", "prefer", "defer")
VALID_PRIORITY_INTS = (1000, 990, 100, 500, 1, -1)
VALID_PRIORITIES = (*VALID_PRIORITY_STRINGS, *VALID_PRIORITY_INTS)
INVALID_PRIORITIES = (0,)

SAMPLE_PINS = (
    "release o=LP-PPA-deadsnakes-ppa-ppa",
    'origin "developer.download.nvidia.com"',
)


@pytest.fixture
def manager(tmp_path):
    yield AptPreferencesManager(path=tmp_path / "preferences")


# region Preference
@pytest.mark.parametrize("priority", VALID_PRIORITIES)
@pytest.mark.parametrize("pin", SAMPLE_PINS)
def test_create_valid_preferences(pin, priority):
    Preference(pin=pin, priority=priority)


@pytest.mark.parametrize("priority", INVALID_PRIORITIES)
def test_invalid_priorities(priority):
    with pytest.raises(AptPreferencesError):
        Preference(pin="", priority=priority)


@pytest.mark.parametrize(
    "input_str,expected",
    [
        (
            dedent(
                """
                Package: *
                Pin: release o=LP-PPA-ppa-ppa
                Pin-Priority: 123
                """
            ),
            Preference(pin="release o=LP-PPA-ppa-ppa", priority=123),
        ),
        (
            dedent(
                """
                Package: *
                Pin: origin "developer.download.nvidia.com"
                Pin-Priority: 456
                """
            ),
            Preference(pin='origin "developer.download.nvidia.com"', priority=456),
        ),
        (
            dedent(
                """
                Explanation: This line will be ignored, but logged.
                Package: *
                Pin: origin "valvesoftware.com"
                Pin-Priority: 789
                """
            ),
            Preference(pin='origin "valvesoftware.com"', priority=789),
        ),
    ],
)
def test_preference_string_parsing(input_str, expected):
    preference = Preference.from_string(input_str)

    assert preference == expected


@pytest.mark.parametrize(
    "preference,expected",
    [
        (
            Preference(pin="release o=LP-PPA-ppa-ppa", priority=123),
            dedent(
                """\
                Package: *
                Pin: release o=LP-PPA-ppa-ppa
                Pin-Priority: 123

                """
            ),
        ),
        (
            Preference(pin='origin "valvesoftware.com"', priority=789),
            dedent(
                """\
                Package: *
                Pin: origin "valvesoftware.com"
                Pin-Priority: 789

                """
            ),
        ),
    ],
)
def test_preference_to_file(preference, expected):
    assert str(preference) == expected


# endregion
# region AptPreferencesManager
def test_read_nonexistent_file(tmp_path):
    preferences_path = tmp_path / "test-preferences"

    manager = AptPreferencesManager(path=preferences_path)

    assert manager._preferences == []


@pytest.mark.parametrize(
    "pref_file,expected",
    [
        (
            "no_header.preferences",
            [
                Preference(pin="release o=LP-PPA-safety-ppa", priority=99999),
                Preference(pin='origin "apt_ppa.redhat.arch.mac"', priority=-1),
            ],
        ),
        (
            "with_header.preferences",
            [
                Preference(pin="release o=LP-PPA-safety-ppa", priority=99999),
                Preference(pin='origin "apt_ppa.redhat.arch.mac"', priority=-1),
            ],
        ),
        ("empty.preferences", []),
        ("only_comment.preferences", []),
        ("many_blank_lines.preferences", []),
    ],
)
def test_read_existing_preferences(test_data_dir, pref_file, expected):
    pref_path = test_data_dir / pref_file
    manager = AptPreferencesManager(path=pref_path)

    manager.read()

    assert manager._preferences == expected


@pytest.mark.parametrize(
    "pref_file,expected_file",
    [
        (
            "no_header.preferences",
            "expected.preferences",
        ),
        (
            "with_header.preferences",
            "expected.preferences",
        ),
        (
            "expected.preferences",
            "expected.preferences",
        ),
    ],
)
def test_read_and_write_correct(test_data_dir, pref_file, expected_file, tmp_path):
    pref_path = test_data_dir / pref_file
    expected_path = test_data_dir / expected_file
    actual_path = tmp_path / "pref"
    shutil.copyfile(pref_path, actual_path)
    manager = AptPreferencesManager(path=actual_path)

    manager.read()
    manager.write()

    assert actual_path.read_text() == expected_path.read_text()


def test_write_empty_preferences_removes_file(tmp_path):
    file = tmp_path / "pref.preferences"
    file.touch()

    manager = AptPreferencesManager(path=file)

    # Still return true if the file was changed, even if that change was removal.
    assert manager.write()
    assert not file.exists()


@pytest.mark.parametrize(
    "preferences,expected_file",
    [
        pytest.param(
            [  # Preferences
                {"priority": 99999, "pin": "release o=LP-PPA-safety-ppa"},
                {"priority": -1, "pin": 'origin "apt_ppa.redhat.arch.mac"'},
            ],
            "expected.preferences",
            id="basic_file",
        )
    ],
)
def test_preferences_added(test_data_dir, tmp_path, preferences, expected_file):
    expected_path = test_data_dir / expected_file
    actual_path = tmp_path / "preferences"
    manager = AptPreferencesManager(path=actual_path)

    for pref in preferences:
        manager.add(**pref)
    manager.write()

    assert actual_path.read_text() == expected_path.read_text()


def test_preferences_path_for_root():
    assert (
        AptPreferencesManager.preferences_path_for_root() == _DEFAULT_PREFERENCES_FILE
    )
    assert AptPreferencesManager.preferences_path_for_root(Path("/my/root")) == Path(
        "/my/root/etc/apt/preferences.d/craft-archives"
    )


# endregion



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2015-2023 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Manage the host's apt source repository configuration."""

import io
import logging
import pathlib
import subprocess
from pathlib import Path
from typing import List, Optional, cast

import distro

from craft_archives import utils

from . import apt_key_manager, apt_ppa, apt_uca, errors, package_repository

logger = logging.getLogger(__name__)

_DEFAULT_SOURCES_DIRECTORY = Path("/etc/apt/sources.list.d")
_DEFAULT_SIGNED_BY_ROOT = Path("/")


def _construct_deb822_source(  # noqa: PLR0913
    *,
    architectures: Optional[List[str]] = None,
    components: Optional[List[str]] = None,
    formats: Optional[List[str]] = None,
    suites: List[str],
    url: str,
    signed_by: pathlib.Path,
) -> str:
    """Construct deb-822 formatted sources string."""
    with io.StringIO() as deb822:
        if formats:
            type_text = " ".join(formats)
        else:
            type_text = "deb"

        print(f"Types: {type_text}", file=deb822)

        print(f"URIs: {url}", file=deb822)

        suites_text = " ".join(suites)
        print(f"Suites: {suites_text}", file=deb822)

        if components:
            components_text = " ".join(components)
            print(f"Components: {components_text}", file=deb822)

        if architectures:
            arch_text = " ".join(architectures)
        else:
            arch_text = utils.get_host_architecture()

        print(f"Architectures: {arch_text}", file=deb822)

        print(f"Signed-By: {str(signed_by)}", file=deb822)

        return deb822.getvalue()


class AptSourcesManager:
    """Manage apt source configuration in /etc/apt/sources.list.d.

    :param sources_list_d: Path to sources.list.d directory.
    """

    # pylint: disable=too-few-public-methods
    def __init__(
        self,
        *,
        sources_list_d: Optional[Path] = None,
        keyrings_dir: Optional[Path] = None,
        signed_by_root: Optional[Path] = None,
    ) -> None:
        """Create a manager for Apt repository sources listings.

        :param sources_list_d: The path to the directory containing the sources listings.
        :param keyrings_dir: The path to the directory containing the (already installed)
          keyrings.
        :param signed_by_root: The path that should be considered as "system root" when
          filling the "Signed-By" field in the sources listings. Used to configure an
          Apt-based system that will eventually be chrooted into.
        """
        self._sources_list_d = sources_list_d or _DEFAULT_SOURCES_DIRECTORY
        self._keyrings_dir = keyrings_dir or apt_key_manager.KEYRINGS_PATH
        self._signed_by_root = signed_by_root or _DEFAULT_SIGNED_BY_ROOT

    @classmethod
    def sources_path_for_root(cls, root: Optional[Path] = None) -> Path:
        """Get the location for Apt source listings with ``root`` as the system root.

        :param root: The optional system root to consider, or None to assume the standard
          system root "/".
        """
        if root is None:
            return _DEFAULT_SOURCES_DIRECTORY
        return root / "etc/apt/sources.list.d"

    def _install_sources(  # noqa: PLR0913
        self,
        *,
        architectures: Optional[List[str]] = None,
        components: Optional[List[str]] = None,
        formats: Optional[List[str]] = None,
        name: str,
        suites: List[str],
        url: str,
        keyring_path: pathlib.Path,
    ) -> bool:
        """Install sources list configuration.

        Write config to:
        /etc/apt/sources.list.d/craft-<name>.sources

        :returns: True if configuration was changed.
        """
        if keyring_path and not keyring_path.is_file():
            raise errors.AptGPGKeyringError(keyring_path)

        keyring_path = Path("/") / keyring_path.relative_to(self._signed_by_root)

        config = _construct_deb822_source(
            architectures=architectures,
            components=components,
            formats=formats,
            suites=suites,
            url=url,
            signed_by=keyring_path,
        )

        if name not in ["default", "default-security"]:
            name = "craft-" + name

        config_path = self._sources_list_d / f"{name}.sources"
        if config_path.exists() and config_path.read_text() == config:
            # Already installed and matches, nothing to do.
            logger.debug(f"Ignoring unchanged sources: {config_path!s}")
            return False

        config_path.write_text(config)
        logger.debug(f"Installed sources: {config_path!s}")
        return True

    def _install_sources_apt(
        self, *, package_repo: package_repository.PackageRepositoryApt
    ) -> bool:
        """Install repository configuration.

        1) First check to see if package repo is implied path,
           or "bare repository" config.  This is indicated when no
           path, components, or suites are indicated.
        2) If path is specified, convert path to a suite entry,
           ending with "/".

        Relatedly, this assumes all of the error-checking has been
        done already on the package_repository object in a proper
        fashion, but do some sanity checks here anyways.

        :returns: True if source configuration was changed.
        """
        if (
            not package_repo.path
            and not package_repo.components
            and not package_repo.suites
        ):
            suites = ["/"]
        elif package_repo.path:
            # Suites denoting exact path must end with '/'.
            path = package_repo.path
            if not path.endswith("/"):
                path += "/"
            suites = [path]
        elif package_repo.suites:
            suites = package_repo.suites
        else:  # pragma: no cover
            raise RuntimeError("no suites or path")

        name = package_repo.name

        keyring_path = apt_key_manager.get_keyring_path(
            package_repo.key_id, base_path=self._keyrings_dir
        )

        return self._install_sources(
            architectures=package_repo.architectures,
            components=package_repo.components,
            formats=cast(Optional[List[str]], package_repo.formats),
            name=name,
            suites=suites,
            url=package_repo.url,
            keyring_path=keyring_path,
        )

    def _install_sources_ppa(
        self, *, package_repo: package_repository.PackageRepositoryAptPPA
    ) -> bool:
        """Install PPA formatted repository.

        Create a sources list config by:
        - Looking up the codename of the host OS and using it as the "suites"
          entry.
        - Formulate deb URL to point to PPA.
        - Enable only "deb" formats.

        :returns: True if source configuration was changed.
        """
        owner, name = apt_ppa.split_ppa_parts(ppa=package_repo.ppa)
        codename = distro.codename()

        key_id = apt_ppa.get_launchpad_ppa_key_id(ppa=package_repo.ppa)
        keyring_path = apt_key_manager.get_keyring_path(
            key_id, base_path=self._keyrings_dir
        )

        return self._install_sources(
            components=["main"],
            formats=["deb"],
            name=f"ppa-{owner}_{name}",
            suites=[codename],
            url=f"http://ppa.launchpad.net/{owner}/{name}/ubuntu",
            keyring_path=keyring_path,
        )

    def _install_sources_uca(
        self, *, package_repo: package_repository.PackageRepositoryAptUCA
    ) -> bool:
        """Install UCA formatted repository.

        Create a sources list config by:
        - Looking up the codename of the host OS and using it as the "suites"
          entry.
        - Formulate deb URL to point to UCA.
        - Enable only "deb" formats.

        :returns: True if source configuration was changed.
        """
        cloud = package_repo.cloud
        pocket = package_repo.pocket

        codename = distro.codename()
        apt_uca.check_release_compatibility(codename, cloud, pocket)

        key_id = package_repository.UCA_KEY_ID
        keyring_path = apt_key_manager.get_keyring_path(
            key_id, base_path=self._keyrings_dir
        )
        return self._install_sources(
            components=["main"],
            formats=["deb"],
            name=f"cloud-{cloud}",
            suites=[f"{codename}-{pocket}/{cloud}"],
            url=package_repository.UCA_ARCHIVE,
            keyring_path=keyring_path,
        )

    def install_package_repository_sources(
        self,
        *,
        package_repo: package_repository.PackageRepository,
    ) -> bool:
        """Install configured package repositories.

        :param package_repo: Repository to install the source configuration for.

        :returns: True if source configuration was changed.
        """
        logger.debug(f"Processing repo: {package_repo!r}")
        if isinstance(package_repo, package_repository.PackageRepositoryAptPPA):
            return self._install_sources_ppa(package_repo=package_repo)

        if isinstance(package_repo, package_repository.PackageRepositoryAptUCA):
            return self._install_sources_uca(package_repo=package_repo)

        if isinstance(package_repo, package_repository.PackageRepositoryApt):
            changed = self._install_sources_apt(package_repo=package_repo)
            architectures = package_repo.architectures
            if changed and architectures:
                _add_architecture(architectures, root=self._signed_by_root)
            return changed

        raise RuntimeError(f"unhandled package repository: {package_repository!r}")


def _add_architecture(architectures: List[str], root: Path) -> None:
    """Add package repository architecture."""
    for arch in architectures:
        logger.info(f"Add repository architecture: {arch}")
        subprocess.run(
            ["dpkg", "--add-architecture", arch, "--root", str(root)],
            check=True,
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2023 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Basic test configuration for craft-archives."""
from pathlib import Path

import pytest


@pytest.fixture(scope="session")
def test_data_dir() -> Path:
    """test_data directory directly under tests directory."""
    path = Path(__file__).parent / "test_data"
    assert path.is_dir()
    return path


@pytest.fixture(scope="session")
def sample_key_path(test_data_dir) -> Path:
    path = test_data_dir / "FC42E99D.asc"
    assert path.is_file()
    return path


@pytest.fixture(scope="session")
def sample_key_string(sample_key_path) -> str:
    return sample_key_path.read_text()


@pytest.fixture(scope="session")
def sample_key_bytes(sample_key_path) -> bytes:
    return sample_key_path.read_bytes()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for craft_archives."""
import logging
import platform
from dataclasses import dataclass
from pathlib import Path
from typing import Dict

logger = logging.getLogger(__name__)


@dataclass
class OSPlatform:
    """Platform definition for a given host."""

    system: str
    release: str
    machine: str

    def __str__(self) -> str:
        """Return the string representation of an OSPlatform."""
        return f"{self.system}/{self.release} ({self.machine})"


# architecture translations from the platform syntax to the deb/snap syntax
# These two architecture mappings are almost inverses of each other, except one map is
# not reversible (same value for different keys)
_ARCH_TRANSLATIONS_PLATFORM_TO_DEB = {
    "aarch64": "arm64",
    "armv7l": "armhf",
    "i686": "i386",
    "ppc": "powerpc",
    "ppc64le": "ppc64el",
    "x86_64": "amd64",
    "AMD64": "amd64",  # Windows support
    "s390x": "s390x",
    "riscv64": "riscv64",
}


_32BIT_USERSPACE_ARCHITECTURE = {
    "aarch64": "armv7l",
    "armv8l": "armv7l",
    "ppc64le": "ppc",
    "x86_64": "i686",
}


def get_os_platform(
    filepath: Path = Path("/etc/os-release"),
) -> OSPlatform:
    """Determine a system/release combo for an OS using /etc/os-release if available."""
    system = platform.system()
    release = platform.release()
    machine = platform.machine()

    if system == "Linux":
        try:
            with filepath.open("rt", encoding="utf-8") as release_file:
                lines = release_file.readlines()
        except FileNotFoundError:
            logger.debug("Unable to locate 'os-release' file, using default values")
        else:
            os_release: Dict[str, str] = {}
            for line in lines:
                line = line.strip()  # noqa PLW2901 — overwriting outer loop variable
                if not line or line.startswith("#") or "=" not in line:
                    continue
                key, value = line.rstrip().split("=", 1)
                if value[0] == value[-1] and value[0] in ('"', "'"):
                    value = value[1:-1]
                os_release[key] = value
            system = os_release.get("ID", system)
            release = os_release.get("VERSION_ID", release)

    return OSPlatform(system=system, release=release, machine=machine)


def get_host_architecture() -> str:
    """Get host architecture in deb format suitable for base definition."""
    os_platform_machine = get_os_platform().machine

    if platform.architecture()[0] == "32bit":
        userspace = _32BIT_USERSPACE_ARCHITECTURE.get(os_platform_machine)
        if userspace:
            os_platform_machine = userspace

    return _ARCH_TRANSLATIONS_PLATFORM_TO_DEB.get(
        os_platform_machine, os_platform_machine
    )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import http
import urllib.error
from pathlib import Path
from textwrap import dedent
from unittest import mock
from unittest.mock import call, patch

import distro
import pytest
from craft_archives.repo import apt_ppa, apt_sources_manager, errors
from craft_archives.repo.apt_sources_manager import (
    _DEFAULT_SOURCES_DIRECTORY,
    AptSourcesManager,
)
from craft_archives.repo.package_repository import (
    PackageRepositoryApt,
    PackageRepositoryAptPPA,
    PackageRepositoryAptUCA,
)

# pyright: reportGeneralTypeIssues=false


@pytest.fixture(autouse=True)
def mock_apt_ppa_get_signing_key(mocker):
    yield mocker.patch(
        "craft_archives.repo.apt_ppa.get_launchpad_ppa_key_id",
        spec=apt_ppa.get_launchpad_ppa_key_id,
        return_value="FAKE-PPA-SIGNING-KEY",
    )


@pytest.fixture(autouse=True)
def mock_environ_copy(mocker):
    yield mocker.patch("os.environ.copy")


@pytest.fixture(autouse=True)
def mock_host_arch(mocker):
    m = mocker.patch("craft_archives.utils.get_host_architecture")
    m.return_value = "FAKE-HOST-ARCH"

    yield m


@pytest.fixture(autouse=True)
def mock_run(mocker):
    yield mocker.patch("subprocess.run")


@pytest.fixture(autouse=True)
def mock_version_codename(monkeypatch):
    mock_codename = mock.Mock(return_value="FAKE-CODENAME")
    monkeypatch.setattr(distro, "codename", mock_codename)
    yield mock_codename


@pytest.fixture
def apt_sources_mgr(tmp_path):
    sources_list_d = tmp_path / "sources.list.d"
    sources_list_d.mkdir(parents=True)
    keyrings_dir = tmp_path / "keyrings"
    keyrings_dir.mkdir(parents=True)

    yield apt_sources_manager.AptSourcesManager(
        sources_list_d=sources_list_d,
        keyrings_dir=keyrings_dir,
    )


def create_apt_sources_mgr(tmp_path: Path, *, use_signed_by_root: bool):
    signed_by_root = None
    if use_signed_by_root:
        signed_by_root = tmp_path

    sources_list_d = tmp_path / "sources.list.d"
    sources_list_d.mkdir(parents=True)
    keyrings_dir = tmp_path / "keyrings"
    keyrings_dir.mkdir(parents=True)

    return apt_sources_manager.AptSourcesManager(
        sources_list_d=sources_list_d,
        keyrings_dir=keyrings_dir,
        signed_by_root=signed_by_root,
    )


@pytest.mark.parametrize("use_signed_by_root", [False, True])
@pytest.mark.parametrize(
    "package_repo,name,content_template",
    [
        (
            PackageRepositoryApt(
                type="apt",
                architectures=["amd64", "arm64"],
                components=["test-component"],
                formats=["deb", "deb-src"],
                key_id="A" * 40,
                suites=["test-suite1", "test-suite2"],
                url="http://test.url/ubuntu",
            ),
            "craft-http_test_url_ubuntu.sources",
            dedent(
                """\
                Types: deb deb-src
                URIs: http://test.url/ubuntu
                Suites: test-suite1 test-suite2
                Components: test-component
                Architectures: amd64 arm64
                Signed-By: {keyring_path}
                """
            ),
        ),
        (
            PackageRepositoryApt(
                type="apt",
                architectures=["amd64", "arm64"],
                formats=["deb", "deb-src"],
                path="dir/subdir",
                key_id="A" * 40,
                url="http://test.url/ubuntu",
            ),
            "craft-http_test_url_ubuntu.sources",
            dedent(
                """\
                    Types: deb deb-src
                    URIs: http://test.url/ubuntu
                    Suites: dir/subdir/
                    Architectures: amd64 arm64
                    Signed-By: {keyring_path}
                    """
            ),
        ),
        (
            PackageRepositoryAptPPA(type="apt", ppa="test/ppa"),
            "craft-ppa-test_ppa.sources",
            dedent(
                """\
                Types: deb
                URIs: http://ppa.launchpad.net/test/ppa/ubuntu
                Suites: FAKE-CODENAME
                Components: main
                Architectures: FAKE-HOST-ARCH
                Signed-By: {keyring_path}
                """
            ),
        ),
        (
            PackageRepositoryAptUCA(type="apt", cloud="fake-cloud"),
            "craft-cloud-fake-cloud.sources",
            dedent(
                """\
                Types: deb
                URIs: http://ubuntu-cloud.archive.canonical.com/ubuntu
                Suites: FAKE-CODENAME-updates/fake-cloud
                Components: main
                Architectures: FAKE-HOST-ARCH
                Signed-By: {keyring_path}
                """
            ),
        ),
    ],
)
def test_install(
    tmp_path,
    package_repo,
    name,
    content_template,
    use_signed_by_root,
    mocker,
):
    run_mock = mocker.patch("subprocess.run")
    mocker.patch("urllib.request.urlopen")

    apt_sources_mgr = create_apt_sources_mgr(
        tmp_path, use_signed_by_root=use_signed_by_root
    )
    sources_path = apt_sources_mgr._sources_list_d / name

    keyring_path = apt_sources_mgr._keyrings_dir / "craft-AAAAAAAA.gpg"
    keyring_path.touch(exist_ok=True)

    if use_signed_by_root:
        signed_by_path = "/keyrings/craft-AAAAAAAA.gpg"
    else:
        signed_by_path = str(keyring_path)

    content = content_template.format(keyring_path=signed_by_path).encode()
    mock_keyring_path = mocker.patch(
        "craft_archives.repo.apt_key_manager.get_keyring_path"
    )
    mock_keyring_path.return_value = keyring_path

    changed = apt_sources_mgr.install_package_repository_sources(
        package_repo=package_repo
    )

    assert changed is True
    assert sources_path.read_bytes() == content

    if use_signed_by_root:
        expected_root = str(tmp_path)
    else:
        expected_root = "/"

    if isinstance(package_repo, PackageRepositoryApt) and package_repo.architectures:
        assert run_mock.mock_calls == [
            call(
                ["dpkg", "--add-architecture", "amd64", "--root", expected_root],
                check=True,
            ),
            call(
                ["dpkg", "--add-architecture", "arm64", "--root", expected_root],
                check=True,
            ),
        ]
    else:
        assert run_mock.mock_calls == []

    run_mock.reset_mock()

    # Verify a second-run does not incur any changes.
    changed = apt_sources_mgr.install_package_repository_sources(
        package_repo=package_repo
    )

    assert changed is False
    assert sources_path.read_bytes() == content
    assert run_mock.mock_calls == []


def test_install_ppa_invalid(apt_sources_mgr):
    repo = PackageRepositoryAptPPA(type="apt", ppa="ppa-missing-slash")

    with pytest.raises(errors.AptPPAInstallError) as raised:
        apt_sources_mgr.install_package_repository_sources(package_repo=repo)

    assert str(raised.value) == (
        "Failed to install PPA 'ppa-missing-slash': invalid PPA format"
    )


@patch(
    "urllib.request.urlopen",
    side_effect=urllib.error.HTTPError("", http.HTTPStatus.NOT_FOUND, "", {}, None),  # type: ignore
)
def test_install_uca_invalid(urllib, apt_sources_mgr):
    repo = PackageRepositoryAptUCA(type="apt", cloud="FAKE-CLOUD")
    with pytest.raises(errors.AptUCAInstallError) as raised:
        apt_sources_mgr.install_package_repository_sources(package_repo=repo)

    assert str(raised.value) == (
        "Failed to install UCA 'FAKE-CLOUD/updates': not a valid release for 'FAKE-CODENAME'"
    )


class UnvalidatedAptRepo(PackageRepositoryApt):
    """Repository with no validation to use for invalid repositories."""

    def validate(self) -> None:
        pass


def test_install_apt_errors(apt_sources_mgr):
    repo = PackageRepositoryApt(
        type="apt",
        architectures=["amd64"],
        url="https://example.com",
        key_id="A" * 40,
    )
    with pytest.raises(errors.AptGPGKeyringError):
        apt_sources_mgr._install_sources_apt(package_repo=repo)


def test_preferences_path_for_root():
    assert AptSourcesManager.sources_path_for_root() == _DEFAULT_SOURCES_DIRECTORY
    assert AptSourcesManager.sources_path_for_root(Path("/my/root")) == Path(
        "/my/root/etc/apt/sources.list.d"
    )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2019-2023 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Package repository installer."""

import pathlib
from typing import Any, Dict, List, Optional

from . import errors
from .apt_key_manager import AptKeyManager
from .apt_preferences_manager import AptPreferencesManager
from .apt_sources_manager import AptSourcesManager
from .package_repository import (
    PackageRepository,
    PackageRepositoryApt,
    PackageRepositoryAptPPA,
    PackageRepositoryAptUCA,
)


def install(
    project_repositories: List[Dict[str, Any]], *, key_assets: pathlib.Path
) -> bool:
    """Add package repositories to the host system.

    :param project_repositories: A list of package repositories to install.
    :param key_assets: The directory containing repository keys.

    :return: Whether a package list refresh is required.
    """
    return _install_repos(
        project_repositories=project_repositories, key_assets=key_assets
    )


def install_in_root(
    project_repositories: List[Dict[str, Any]],
    root: pathlib.Path,
    *,
    key_assets: pathlib.Path,
) -> bool:
    """Add package repositories to the system located at ``root``.

    :param project_repositories: A list of package repositories to install.
    :param key_assets: The directory containing repository keys.
    :param root: The directory containing the Apt-based system installation.

    :return: Whether a package list refresh is required.
    """
    return _install_repos(
        project_repositories=project_repositories, root=root, key_assets=key_assets
    )


def _install_repos(
    *,
    project_repositories: List[Dict[str, Any]],
    root: Optional[pathlib.Path] = None,
    key_assets: pathlib.Path,
) -> bool:
    keyrings_path = AptKeyManager.keyrings_path_for_root(root)
    key_manager = AptKeyManager(keyrings_path=keyrings_path, key_assets=key_assets)

    sources_list_d = AptSourcesManager.sources_path_for_root(root)
    sources_manager = AptSourcesManager(
        sources_list_d=sources_list_d,
        keyrings_dir=keyrings_path,
        signed_by_root=root,
    )

    preferences_path = AptPreferencesManager.preferences_path_for_root(root)
    preferences_manager = AptPreferencesManager(path=preferences_path)

    package_repositories = _unmarshal_repositories(project_repositories)

    refresh_required = False
    for package_repo in package_repositories:
        refresh_required |= key_manager.install_package_repository_key(
            package_repo=package_repo
        )
        refresh_required |= sources_manager.install_package_repository_sources(
            package_repo=package_repo
        )
        if (
            isinstance(
                package_repo,
                (
                    PackageRepositoryApt,
                    PackageRepositoryAptPPA,
                    PackageRepositoryAptUCA,
                ),
            )
            and package_repo.priority is not None
        ):
            refresh_required |= preferences_manager.add(
                pin=package_repo.pin, priority=int(package_repo.priority)
            )

    refresh_required |= preferences_manager.write()

    _verify_all_key_assets_installed(key_assets=key_assets, key_manager=key_manager)

    return refresh_required


def _verify_all_key_assets_installed(
    *,
    key_assets: pathlib.Path,
    key_manager: AptKeyManager,
) -> None:
    """Verify all configured key assets are utilized, error if not."""
    for key_asset in key_assets.glob("*"):
        key = key_asset.read_text()
        for key_id in key_manager.get_key_fingerprints(key=key):
            if not key_manager.is_key_installed(key_id=key_id):
                raise errors.PackageRepositoryError(
                    "Found unused key asset {key_asset!r}.",
                    details="All configured key assets must be utilized.",
                    resolution="Verify key usage and remove all unused keys.",
                )


def _unmarshal_repositories(
    project_repositories: List[Dict[str, Any]]
) -> List[PackageRepository]:
    """Create package repositories objects from project data."""
    repositories: List[PackageRepository] = []
    for data in project_repositories:
        pkg_repo: PackageRepository

        if "ppa" in data:
            pkg_repo = PackageRepositoryAptPPA.unmarshal(data)
        elif "cloud" in data:
            pkg_repo = PackageRepositoryAptUCA.unmarshal(data)
        else:
            pkg_repo = PackageRepositoryApt.unmarshal(data)

        repositories.append(pkg_repo)

    return repositories



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2022-2023 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pydantic
import pytest
from craft_archives.repo.projects import validate_repository


# region validate_repository tests
@pytest.mark.parametrize(
    "repo",
    [
        {"type": "apt", "ppa": "ppa/ppa"},
        {
            "type": "apt",
            "url": "https://deb.repo",
            "key-id": "A" * 40,
            "formats": ["deb"],
        },
        {"type": "apt", "cloud": "antelope"},
    ],
)
def test_validate_repository(repo):
    validate_repository(repo)


def test_validate_repository_invalid():
    with pytest.raises(TypeError, match="must be a dictionary"):
        validate_repository("invalid repository")  # type: ignore


def test_validate_repository_empty_dict():
    with pytest.raises(pydantic.ValidationError):
        validate_repository({})


# endregion



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2023 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Manage the host's apt preferences configurations."""
import contextlib
import dataclasses
import io
import logging
import typing
from pathlib import Path

from craft_archives.repo.errors import AptPreferencesError

_DEFAULT_PREFERENCES_FILE = Path("/etc/apt/preferences.d/craft-archives")

_DEFAULT_HEADER = "# This file is managed by craft-archives"

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class Preference:
    """A single preference paragraph.

    Matches preferences as defined by apt:
    https://manpages.debian.org/bullseye/apt/apt_preferences.5.en.html
    """

    pin: str
    priority: int

    @classmethod
    def from_string(cls, input_str: str) -> "Preference":
        """Create a Preference object from a preferences compliant string.

        :param input_str: The string containing a preferences paragraph
        :returns: a single Preference object
        :raises: ValueError if the string is not a preferences paragraph
        :raises: AptPreferencesError if either pin or priority is missing
        """
        lines = input_str.splitlines(keepends=False)
        pin, priority = None, None
        for line in lines:
            if not line or line.startswith("#"):
                continue
            key, value = line.split(":", maxsplit=1)
            key, value = key.strip().lower(), value.strip()
            if key == "pin":
                pin = value
            elif key == "pin-priority":
                priority = int(value)
            else:
                logger.warning(f"Unknown preference line: {line!r}")
        if pin is None and priority is None:
            raise ValueError("String is not a preferences paragraph.")
        if not pin:
            raise AptPreferencesError(
                component="pin",
                value=pin,
                resolution="Remove or update preferences file.",
            )
        if not priority:
            raise AptPreferencesError(
                component="priority",
                value=priority,
                resolution="Remove or update preferences file.",
            )
        return cls(pin=pin, priority=priority)

    def __post_init__(self) -> None:
        if self.priority == 0:
            raise AptPreferencesError(
                component="pin",
                details="Pin-Priority cannot be zero.",
                resolution="Check pin values for repositories.",
            )

    def __str__(self) -> str:
        """Return the preference paragraph as a string."""
        with io.StringIO() as file:
            print("Package: *", file=file)
            print(f"Pin: {self.pin}", file=file)
            print(f"Pin-Priority: {self.priority}", file=file)
            print("", file=file)  # Empty line to start a new paragraph.
            return file.getvalue()


class AptPreferencesManager:
    """Manage an apt preferences file.

    :param path: Path to the preferences file
    :param header: A comment string to act as the file header.
    """

    def __init__(
        self,
        *,
        path: typing.Optional[Path] = None,
        header: str = _DEFAULT_HEADER,
    ) -> None:
        self._header = header
        self._path = path or _DEFAULT_PREFERENCES_FILE
        self._preferences: typing.List[Preference] = []

    @classmethod
    def preferences_path_for_root(cls, root: typing.Optional[Path] = None) -> Path:
        """Get the location for the Apt preferences file with ``root`` as the system root.

        :param root: The optional system root to consider, or None to assume the standard
          system root "/".
        """
        if root is None:
            return _DEFAULT_PREFERENCES_FILE
        return root / "etc/apt/preferences.d/craft-archives"

    def read(self) -> None:
        """Read the preferences file and populate Preferences objects."""
        if not self._path.exists():
            logger.debug("No preferences read.")
            return
        paragraphs = self._path.read_text().split("\n\n")
        for paragraph in paragraphs:
            # Strip and check for contents to ensure no empty paragraphs are included.
            if not paragraph.strip():
                continue
            with contextlib.suppress(ValueError):
                preference = Preference.from_string(paragraph)
                if preference not in self._preferences:
                    self._preferences.append(preference)
        logger.debug(f"{len(self._preferences)} pin preferences read.")

    def write(self) -> bool:
        """Write the preferences file from the given preferences.

        :returns: True if the preferences file was changed.
        """
        if not self._preferences:
            if self._path.exists():
                self._path.unlink()
                return True
            return False
        with io.StringIO() as config:
            print(self._header, file=config)
            for preference in self._preferences:
                config.write(str(preference))

            config_str = config.getvalue()

        if self._path.exists() and self._path.read_text() == config_str:
            logger.debug(f"Ignoring unchanged preferences: {self._path!s}")
            return False

        self._path.write_text(config_str)
        return True

    def add(self, *, pin: str, priority: int) -> bool:
        """Add a preference.

        :returns: True if the preference was added, False if already existed.
        """
        preference = Preference(pin, priority)
        if preference in self._preferences:
            return False
        self._preferences.append(preference)
        return True



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2019-2023 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Package repository definitions."""

import abc
import enum
import re
from typing import Any, Dict, List, Mapping, Optional, Union
from urllib.parse import urlparse

import pydantic
from overrides import overrides  # pyright: ignore[reportUnknownVariableType]
from pydantic import (
    AnyUrl,
    ConstrainedStr,
    root_validator,  # pyright: ignore[reportUnknownVariableType]
    validator,  # pyright: ignore[reportUnknownVariableType]
)

# NOTE: using this instead of typing.Literal because of this bad typing_extensions
# interaction: https://github.com/pydantic/pydantic/issues/5821#issuecomment-1559196859
# We can revisit this when typing_extensions >4.6.0 is released, and/or we no longer
# have to support Python <3.10
from typing_extensions import Literal

from . import errors

UCA_ARCHIVE = "http://ubuntu-cloud.archive.canonical.com/ubuntu"
UCA_NETLOC = urlparse(UCA_ARCHIVE).netloc
UCA_VALID_POCKETS = ["updates", "proposed"]
UCA_DEFAULT_POCKET = UCA_VALID_POCKETS[0]
UCA_KEY_ID = "391A9AA2147192839E9DB0315EDB1B62EC4926EA"


class KeyIdStr(ConstrainedStr):
    """A constrained string for a GPG key ID."""

    min_length = 40
    max_length = 40
    regex = re.compile(r"^[0-9A-F]{40}$")


class PriorityString(enum.IntEnum):
    """Convenience values that represent common deb priorities."""

    ALWAYS = 1000
    PREFER = 990
    DEFER = 100


PriorityValue = Union[int, Literal["always", "prefer", "defer"]]


def _alias_generator(value: str) -> str:
    return value.replace("_", "-")


class PackageRepository(pydantic.BaseModel, abc.ABC):
    """The base class for package repositories."""

    class Config:  # pylint: disable=too-few-public-methods
        """Pydantic model configuration."""

        validate_assignment = True
        allow_mutation = False
        allow_population_by_field_name = True
        alias_generator = _alias_generator
        extra = "forbid"

    type: Literal["apt"]
    priority: Optional[PriorityValue]

    @root_validator
    def priority_cannot_be_zero(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Priority cannot be zero per apt Preferences specification."""
        priority = values.get("priority")
        if priority == 0:
            raise _create_validation_error(
                url=str(values.get("url") or values.get("ppa") or values.get("cloud")),
                message="invalid priority: Priority cannot be zero.",
            )
        return values

    @validator("priority")
    def _convert_priority_to_int(
        cls, priority: Optional[PriorityValue], values: Dict[str, Any]
    ) -> Optional[int]:
        if isinstance(priority, str):
            str_priority = priority.upper()
            if str_priority in PriorityString.__members__:
                return PriorityString[str_priority]
            # This cannot happen; if it's a string but not one of the accepted
            # ones Pydantic will fail early and won't call this validator.
            raise _create_validation_error(
                url=str(values.get("url") or values.get("ppa") or values.get("cloud")),
                message=(
                    f"invalid priority {priority!r}. "
                    "Priority must be 'always', 'prefer', 'defer' or a nonzero integer."
                ),
            )
        return priority

    def marshal(self) -> Dict[str, Union[str, int]]:
        """Return the package repository data as a dictionary."""
        return self.dict(by_alias=True, exclude_none=True)

    @classmethod
    def unmarshal(cls, data: Mapping[str, Any]) -> "PackageRepository":
        """Create a package repository object from the given data."""
        if not isinstance(data, dict):  # pyright: ignore[reportUnnecessaryIsInstance]
            raise errors.PackageRepositoryValidationError(
                url=str(data),
                brief="invalid object.",
                details="Package repository must be a valid dictionary object.",
                resolution=(
                    "Verify repository configuration and ensure that the "
                    "correct syntax is used."
                ),
            )

        if "ppa" in data:
            return PackageRepositoryAptPPA.unmarshal(data)
        if "cloud" in data:
            return PackageRepositoryAptUCA.unmarshal(data)

        return PackageRepositoryApt.unmarshal(data)

    @classmethod
    def unmarshal_package_repositories(
        cls, data: Optional[List[Dict[str, Any]]]
    ) -> List["PackageRepository"]:
        """Create multiple package repositories from the given data."""
        repositories: List[PackageRepository] = []

        if data is not None:
            if not isinstance(
                data, list
            ):  # pyright: ignore[reportUnnecessaryIsInstance]
                raise errors.PackageRepositoryValidationError(
                    url=str(data),
                    brief="invalid list object.",
                    details="Package repositories must be a list of objects.",
                    resolution=(
                        "Verify 'package-repositories' configuration and ensure "
                        "that the correct syntax is used."
                    ),
                )

            for repository in data:
                package_repo = cls.unmarshal(repository)
                repositories.append(package_repo)

        return repositories


class PackageRepositoryAptPPA(PackageRepository):
    """A PPA package repository."""

    ppa: str

    @validator("ppa")
    def _non_empty_ppa(cls, ppa: str) -> str:
        if not ppa:
            raise _create_validation_error(
                message="Invalid PPA: PPAs must be non-empty strings."
            )
        return ppa

    @classmethod
    @overrides
    def unmarshal(cls, data: Mapping[str, Any]) -> "PackageRepositoryAptPPA":
        """Create a package repository object from the given data."""
        return cls(**data)

    @property
    def pin(self) -> str:
        """The pin string for this repository if needed."""
        ppa_origin = self.ppa.replace("/", "-")
        return f"release o=LP-PPA-{ppa_origin}"


class PackageRepositoryAptUCA(PackageRepository):
    """A cloud package repository."""

    cloud: str
    pocket: Literal["updates", "proposed"] = "updates"

    @validator("cloud")
    def _non_empty_cloud(cls, cloud: str) -> str:
        if not cloud:
            raise _create_validation_error(message="clouds must be non-empty strings.")
        return cloud

    @classmethod
    @overrides
    def unmarshal(cls, data: Mapping[str, Any]) -> "PackageRepositoryAptUCA":
        """Create a package repository object from the given data."""
        return cls(**data)

    @property
    def pin(self) -> str:
        """The pin string for this repository if needed."""
        return f'origin "{UCA_NETLOC}"'


class PackageRepositoryApt(PackageRepository):
    """An APT package repository."""

    url: AnyUrl
    key_id: KeyIdStr = pydantic.Field(alias="key-id")
    architectures: Optional[List[str]]
    formats: Optional[List[Literal["deb", "deb-src"]]]
    path: Optional[str]
    components: Optional[List[str]]
    key_server: Optional[str] = pydantic.Field(alias="key-server")
    suites: Optional[List[str]]

    # Customize some of the validation error messages
    class Config(PackageRepository.Config):  # noqa: D106 - no docstring needed
        error_msg_templates = {
            "value_error.any_str.min_length": "Invalid URL; URLs must be non-empty strings"
        }

    @property
    def name(self) -> str:
        """Get the repository name."""
        return re.sub(r"\W+", "_", self.url)

    @validator("path")
    def _path_non_empty(
        cls, path: Optional[str], values: Dict[str, Any]
    ) -> Optional[str]:
        if path is not None and not path:
            raise _create_validation_error(
                url=values.get("url"),
                message="Invalid path; Paths must be non-empty strings.",
            )
        return path

    @validator("components")
    def _not_mixing_components_and_path(
        cls, components: Optional[List[str]], values: Dict[str, Any]
    ) -> Optional[List[str]]:
        path = values.get("path")
        if components and path:
            raise _create_validation_error(
                url=values.get("url"),
                message=(
                    f"components {components!r} cannot be combined with "
                    f"path {path!r}."
                ),
            )
        return components

    @validator("suites")
    def _not_mixing_suites_and_path(
        cls, suites: Optional[List[str]], values: Dict[str, Any]
    ) -> Optional[List[str]]:
        path = values.get("path")
        if suites and path:
            message = f"suites {suites!r} cannot be combined with path {path!r}."
            raise _create_validation_error(url=values.get("url"), message=message)
        return suites

    @validator("suites", each_item=True)
    def _suites_without_backslash(cls, suite: str, values: Dict[str, Any]) -> str:
        if suite.endswith("/"):
            raise _create_validation_error(
                url=values.get("url"),
                message=f"invalid suite {suite!r}. Suites must not end with a '/'.",
            )
        return suite

    @root_validator
    def _missing_components_or_suites(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        suites = values.get("suites")
        components = values.get("components")
        url = values.get("url")
        if suites and not components:
            raise _create_validation_error(
                url=url, message="components must be specified when using suites."
            )
        if components and not suites:
            raise _create_validation_error(
                url=url, message="suites must be specified when using components."
            )

        return values

    @classmethod
    @overrides
    def unmarshal(cls, data: Mapping[str, Any]) -> "PackageRepositoryApt":
        """Create a package repository object from the given data."""
        return cls(**data)

    @property
    def pin(self) -> str:
        """The pin string for this repository if needed."""
        domain = urlparse(self.url).netloc
        return f'origin "{domain}"'


def _create_validation_error(*, url: Optional[str] = None, message: str) -> ValueError:
    """Create a ValueError with a formatted message and an optional indicative ``url``."""
    error_message = ""
    if url:
        error_message += f"Invalid package repository for '{url}': "
    error_message += message
    return ValueError(error_message)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2023 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Integration tests for repo.installer"""

import shutil
from pathlib import Path
from textwrap import dedent
from typing import Any, Dict, List

import distro
import pytest
from craft_archives import repo

APT_SOURCES = dedent(
    """
    Types: deb deb-src
    URIs: http://ppa.launchpad.net/snappy-dev/snapcraft-daily/ubuntu
    Suites: focal
    Components: main
    Architectures: amd64
    Signed-By: {key_location}/craft-FC42E99D.gpg
    """
).lstrip()

VERSION_CODENAME = distro.codename()

PPA_SOURCES = dedent(
    """
    Types: deb
    URIs: http://ppa.launchpad.net/deadsnakes/ppa/ubuntu
    Suites: {codename}
    Components: main
    Architectures: amd64
    Signed-By: {key_location}/craft-6A755776.gpg
    """
).lstrip()

# Needed because some "clouds" are only available for specific Ubuntu releases
RELEASE_TO_CLOUD = {
    "jammy": {"cloud": "antelope", "codename": "jammy"},
    "focal": {"cloud": "wallaby", "codename": "focal"},
}

CLOUD_DATA = RELEASE_TO_CLOUD[VERSION_CODENAME]

CLOUD_SOURCES = dedent(
    """
    Types: deb
    URIs: http://ubuntu-cloud.archive.canonical.com/ubuntu
    Suites: {codename}-updates/{cloud}
    Components: main
    Architectures: amd64
    Signed-By: {key_location}/craft-EC4926EA.gpg
    """
).lstrip()

PREFERENCES = dedent(
    """
    # This file is managed by craft-archives
    Package: *
    Pin: origin "ppa.launchpad.net"
    Pin-Priority: 100

    Package: *
    Pin: release o=LP-PPA-deadsnakes-ppa
    Pin-Priority: 1000

    Package: *
    Pin: origin "ubuntu-cloud.archive.canonical.com"
    Pin-Priority: 123

    """
).lstrip()


def create_etc_apt_dirs(etc_apt: Path):
    etc_apt.mkdir(parents=True)

    keyrings_dir = etc_apt / "keyrings"
    keyrings_dir.mkdir()

    sources_dir = etc_apt / "sources.list.d"
    sources_dir.mkdir()

    preferences_dir = etc_apt / "preferences.d"
    preferences_dir.mkdir()


@pytest.fixture
def fake_etc_apt(tmp_path, mocker) -> Path:
    """Mock the default paths used to store keys, sources and preferences."""
    etc_apt = tmp_path / "etc/apt"
    create_etc_apt_dirs(etc_apt)

    keyrings_dir = etc_apt / "keyrings"
    mocker.patch("craft_archives.repo.apt_key_manager.KEYRINGS_PATH", new=keyrings_dir)

    sources_dir = etc_apt / "sources.list.d"
    mocker.patch(
        "craft_archives.repo.apt_sources_manager._DEFAULT_SOURCES_DIRECTORY",
        new=sources_dir,
    )

    preferences_dir = etc_apt / "preferences.d"
    preferences_dir = preferences_dir / "craft-archives"
    mocker.patch(
        "craft_archives.repo.apt_preferences_manager._DEFAULT_PREFERENCES_FILE",
        new=preferences_dir,
    )

    return etc_apt


@pytest.fixture()
def all_repo_types() -> List[Dict[str, Any]]:
    return [
        # a "standard" repo, with a key coming from the assets dir
        {
            "type": "apt",
            "formats": ["deb", "deb-src"],
            "components": ["main"],
            "suites": ["focal"],
            "key-id": "78E1918602959B9C59103100F1831DDAFC42E99D",
            "url": "http://ppa.launchpad.net/snappy-dev/snapcraft-daily/ubuntu",
            "priority": "defer",
        },
        # a "ppa" repo, with key coming from the ubuntu keyserver
        {
            "type": "apt",
            "ppa": "deadsnakes/ppa",
            "priority": "always",
        },
        # a "cloud" repo
        {
            "type": "apt",
            "cloud": CLOUD_DATA["cloud"],
            "pocket": "updates",
            "priority": 123,
        },
    ]


@pytest.fixture
def test_keys_dir(tmp_path, sample_key_path) -> Path:
    target_dir = tmp_path / "keys"
    target_dir.mkdir()

    shutil.copy2(sample_key_path, target_dir)

    return target_dir


def test_install(fake_etc_apt, all_repo_types, test_keys_dir):
    """Integrated test that checks the configuration of keys, sources and pins."""

    assert repo.install(project_repositories=all_repo_types, key_assets=test_keys_dir)

    check_keyrings(fake_etc_apt)
    check_sources(fake_etc_apt, signed_by_location=fake_etc_apt / "keyrings")
    check_preferences(fake_etc_apt)


def test_install_in_root(tmp_path, all_repo_types, test_keys_dir):
    """Integrated test that checks the configuration of keys, sources and pins."""
    etc_apt = tmp_path / "etc/apt"
    create_etc_apt_dirs(etc_apt)

    assert repo.install_in_root(
        project_repositories=all_repo_types, key_assets=test_keys_dir, root=tmp_path
    )

    check_keyrings(etc_apt)
    check_sources(etc_apt, signed_by_location=Path("/etc/apt/keyrings"))
    check_preferences(etc_apt)


def check_keyrings(etc_apt_dir: Path) -> None:
    keyrings_dir = etc_apt_dir / "keyrings"

    # Must have exactly these keyring files, one for each repo
    expected_key_ids = ("6A755776", "FC42E99D", "EC4926EA")

    assert len(list(keyrings_dir.iterdir())) == len(expected_key_ids)
    for key_id in expected_key_ids:
        keyring_file = keyrings_dir / f"craft-{key_id}.gpg"
        assert keyring_file.is_file()


def check_sources(etc_apt_dir: Path, signed_by_location: Path) -> None:
    sources_dir = etc_apt_dir / "sources.list.d"

    keyrings_on_fs = etc_apt_dir / "keyrings"

    cloud_name = CLOUD_DATA["cloud"]
    codename = CLOUD_DATA["codename"]

    # Must have exactly these sources files, one for each repo
    source_to_contents = {
        "http_ppa_launchpad_net_snappy_dev_snapcraft_daily_ubuntu": APT_SOURCES.format(
            key_location=signed_by_location
        ),
        "ppa-deadsnakes_ppa": PPA_SOURCES.format(
            codename=VERSION_CODENAME, key_location=signed_by_location
        ),
        f"cloud-{cloud_name}": CLOUD_SOURCES.format(
            cloud=cloud_name,
            codename=codename,
            key_location=signed_by_location,
        ),
    }

    assert len(list(keyrings_on_fs.iterdir())) == len(source_to_contents)

    for source_repo, expected_contents in source_to_contents.items():
        source_file = sources_dir / f"craft-{source_repo}.sources"
        assert source_file.is_file()
        assert source_file.read_text() == expected_contents


def check_preferences(etc_apt_dir: Path) -> None:
    # Exactly one "preferences" file
    preferences_dir = etc_apt_dir / "preferences.d"
    assert len(list(preferences_dir.iterdir())) == 1

    preferences_file = preferences_dir / "craft-archives"
    assert preferences_file.is_file()
    assert preferences_file.read_text() == PREFERENCES



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2020-2023 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import http
import urllib.error
from unittest.mock import Mock, patch

import pytest
from craft_archives.repo import apt_uca, errors


@patch("urllib.request.urlopen", return_value=Mock(status=http.HTTPStatus.OK))
def test_check_release_compatibility(urllib):
    assert apt_uca.check_release_compatibility("jammy", "antelope") is None


@patch("urllib.request.urlopen", side_effect=urllib.error.HTTPError("", http.HTTPStatus.NOT_FOUND, "NOT FOUND", {}, None))  # type: ignore
def test_check_release_compatibility_invalid(urllib):
    with pytest.raises(
        errors.AptUCAInstallError,
        match="Failed to install UCA 'invalid-cloud/updates': not a valid release for 'jammy'",
    ):
        apt_uca.check_release_compatibility("jammy", "invalid-cloud")


@patch("urllib.request.urlopen", side_effect=urllib.error.HTTPError("", http.HTTPStatus.BAD_GATEWAY, "BAD GATEWAY", {}, None))  # type: ignore
def test_check_release_compatibility_bad_gateway(urllib):
    with pytest.raises(
        errors.AptUCAInstallError,
        match="Failed to install UCA 'antelope/updates': unexpected status code 502: 'BAD GATEWAY' while fetching release",
    ):
        apt_uca.check_release_compatibility("jammy", "antelope")



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2022-2023 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Package repository error definitions."""
import pathlib
from typing import Any, Literal, Optional

from craft_archives.errors import ArchivesError


class PackageRepositoryError(ArchivesError):
    """Package repository error base."""


class PackageRepositoryValidationError(PackageRepositoryError):
    """Package repository is invalid."""

    def __init__(
        self,
        url: str,
        brief: str,
        details: Optional[str] = None,
        resolution: Optional[str] = None,
    ) -> None:
        super().__init__(
            f"Invalid package repository for {url!r}: {brief}",
            details=details,
            resolution=resolution,
        )


class AptPreferencesError(PackageRepositoryError):
    """Apt preferences are invalid."""

    def __init__(
        self,
        component: Literal["pin", "priority"],
        value: Optional[Any] = None,
        details: Optional[str] = None,
        resolution: Optional[str] = None,
    ) -> None:
        super().__init__(
            brief=f"Invalid repository preference {component}",
            details=details or f"Invalid value: {value!r}",
            resolution=resolution,
        )


class AptPPAInstallError(PackageRepositoryError):
    """Installation of a PPA repository failed."""

    def __init__(self, ppa: str, reason: str) -> None:
        super().__init__(
            f"Failed to install PPA {ppa!r}: {reason}",
            resolution="Verify PPA is correct and try again",
        )


class AptUCAInstallError(PackageRepositoryError):
    """Installation of an UCA repository failed."""

    def __init__(self, cloud: str, pocket: str, reason: str) -> None:
        super().__init__(
            f"Failed to install UCA '{cloud}/{pocket}': {reason}",
            resolution="Verify UCA is correct and try again",
        )


class AptGPGKeyringError(PackageRepositoryError):
    """GPG keyring for repository does not exist or not valid."""

    def __init__(self, keyring_path: pathlib.Path) -> None:
        super().__init__(
            "Could not find keyring file for repository.",
            f"Keyring file does not exist or is invalid: {keyring_path}",
            "Ensure the keyring is installed in the correct path.",
        )


class AptGPGKeyInstallError(PackageRepositoryError):
    """Installation of GPG key failed."""

    def __init__(
        self,
        output: str,
        *,
        key: Optional[str] = None,
        key_id: Optional[str] = None,
        key_server: Optional[str] = None,
    ) -> None:
        """Convert apt-key's output into a more user-friendly message."""
        message = output.replace(
            "Warning: apt-key output should not be parsed (stdout is not a terminal)",
            "",
        ).strip()

        # Improve error messages that we can.
        if (
            "gpg: keyserver receive failed: No data" in message
            and key_id
            and key_server
        ):
            message = f"GPG key {key_id!r} not found on key server {key_server!r}"
        elif (
            "gpg: keyserver receive failed: Server indicated a failure" in message
            and key_server
        ):
            message = f"unable to establish connection to key server {key_server!r}"
        elif (
            "gpg: keyserver receive failed: Connection timed out" in message
            and key_server
        ):
            message = (
                f"unable to establish connection to key server {key_server!r} "
                f"(connection timed out)"
            )

        details = ""
        if key:
            details += f"GPG key:\n{key}\n"
        if key_id:
            details += f"GPG key ID: {key_id}\n"
        if key_server:
            details += f"GPG key server: {key_server}"

        super().__init__(
            f"Failed to install GPG key: {message}",
            details=details,
            resolution="Verify any configured GPG keys",
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2022-2023 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import pytest
from craft_archives.repo import installer
from craft_archives.repo.apt_key_manager import AptKeyManager
from craft_archives.repo.apt_preferences_manager import AptPreferencesManager
from craft_archives.repo.apt_sources_manager import AptSourcesManager
from craft_archives.repo.package_repository import (
    PackageRepositoryApt,
    PackageRepositoryAptPPA,
    PackageRepositoryAptUCA,
)

# ruff: noqa: PLR2004


def test_unmarshal_repositories():
    data = [
        {
            "type": "apt",
            "ppa": "test/somerepo",
        },
        {
            "type": "apt",
            "url": "https://some/url",
            "key-id": "ABCDE12345" * 4,
        },
        {
            "type": "apt",
            "ppa": "test/somerepo",
            "priority": -1,
        },
        {
            "type": "apt",
            "url": "https://some/url",
            "key-id": "ABCDE12345" * 4,
            "priority": -2,
        },
        {
            "type": "apt",
            "cloud": "antelope",
            "pocket": "proposed",
        },
    ]

    pkg_repos = installer._unmarshal_repositories(data)
    assert len(pkg_repos) == 5
    assert isinstance(pkg_repos[0], PackageRepositoryAptPPA)
    assert pkg_repos[0].ppa == "test/somerepo"
    assert pkg_repos[0].priority is None
    assert isinstance(pkg_repos[1], PackageRepositoryApt)
    assert pkg_repos[1].url == "https://some/url"
    assert pkg_repos[1].key_id == "ABCDE12345" * 4
    assert pkg_repos[1].priority is None
    assert isinstance(pkg_repos[2], PackageRepositoryAptPPA)
    assert isinstance(pkg_repos[3], PackageRepositoryApt)
    assert pkg_repos[2].priority == -1
    assert pkg_repos[3].priority == -2
    assert isinstance(pkg_repos[4], PackageRepositoryAptUCA)
    assert pkg_repos[4].cloud == "antelope"
    assert pkg_repos[4].pocket == "proposed"


@pytest.fixture
def manager_mocks(mocker, tmp_path):
    mock_install_package_repository_key = mocker.patch.object(
        AptKeyManager,
        "install_package_repository_key",
        return_value=True,
    )
    mock_install_package_repository_sources = mocker.patch.object(
        AptSourcesManager,
        "install_package_repository_sources",
        return_value=True,
    )
    mock_preferences_add = mocker.patch.object(
        AptPreferencesManager,
        "add",
    )
    mock_preferences_write = mocker.patch.object(
        AptPreferencesManager,
        "write",
        return_value=True,
    )

    repo_dict = {
        "type": "apt",
        "ppa": "test/somerepo",
        "priority": 999,
    }
    repo_ppa = PackageRepositoryAptPPA.unmarshal(repo_dict)
    yield [repo_dict]

    mock_install_package_repository_key.assert_called_once_with(package_repo=repo_ppa)
    mock_install_package_repository_sources.assert_called_once_with(
        package_repo=repo_ppa
    )
    mock_preferences_add.assert_called_once_with(
        pin="release o=LP-PPA-test-somerepo", priority=999
    )
    assert mock_preferences_write.called


def test_install(tmp_path, manager_mocks):
    """Smokish test that checks that installer.install() makes the expected calls."""
    project_repositories = manager_mocks
    assert installer.install(
        project_repositories=project_repositories, key_assets=tmp_path
    )


def test_install_in_root(tmp_path, manager_mocks):
    """Smokish test that checks that installer.install_in_root() makes the expected calls."""
    project_repositories = manager_mocks
    assert installer.install_in_root(
        project_repositories=project_repositories, key_assets=tmp_path, root=tmp_path
    )



# This file is part of craft-archives.
#
# Copyright 2023 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""craft-archives package demo."""

try:
    from ._version import __version__
except ImportError:  # pragma: no cover
    from importlib.metadata import version, PackageNotFoundError

    try:
        __version__ = version("craft-archives")
    except PackageNotFoundError:
        __version__ = "dev"

__all__ = [
    "__version__",
]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2020-2023 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Ubuntu Cloud Archive helpers."""

import http
import urllib.error
import urllib.parse
import urllib.request

from . import errors
from .package_repository import (
    UCA_ARCHIVE,
    UCA_DEFAULT_POCKET,
)


def check_release_compatibility(
    codename: str, cloud: str, pocket: str = UCA_DEFAULT_POCKET
) -> None:
    """Raise an exception if the release is incompatible with codename."""
    request = UCA_ARCHIVE + f"/dists/{codename}-{pocket}/{cloud}/"
    try:
        urllib.request.urlopen(request)
    except urllib.error.HTTPError as e:
        if e.code == http.HTTPStatus.NOT_FOUND:
            raise errors.AptUCAInstallError(
                cloud, pocket, f"not a valid release for {codename!r}"
            )
        raise errors.AptUCAInstallError(
            cloud,
            pocket,
            f"unexpected status code {e.code}: {e.reason!r} while fetching release",
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2020-2023 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import subprocess
from pathlib import Path
from unittest import mock
from unittest.mock import call

import pytest
from craft_archives.repo import apt_ppa, errors, package_repository
from craft_archives.repo.apt_key_manager import KEYRINGS_PATH, AptKeyManager
from craft_archives.repo.package_repository import (
    PackageRepositoryApt,
    PackageRepositoryAptPPA,
    PackageRepositoryAptUCA,
)

# pyright: reportGeneralTypeIssues=false

SAMPLE_GPG_SHOW_KEY_OUTPUT = b"""\
pub:-:4096:1:F1831DDAFC42E99D:1416490823:::-:::scSC::::::23::0:
fpr:::::::::FAKE-KEY-ID-FROM-GNUPG:
uid:-::::1416490823::DCB9EEE37DC9FD84C3DB333BFBF6C41A075EEF62::Launchpad PPA for Snappy Developers::::::::::0:
"""


@pytest.fixture(autouse=True)
def mock_environ_copy(mocker):
    yield mocker.patch("os.environ.copy")


@pytest.fixture(autouse=True)
def mock_run(mocker):
    yield mocker.patch("subprocess.run", spec=subprocess.run)


@pytest.fixture
def mock_chmod(mocker):
    return mocker.patch("pathlib.Path.chmod", autospec=True)


@pytest.fixture(autouse=True)
def mock_apt_ppa_get_signing_key(mocker):
    yield mocker.patch(
        "craft_archives.repo.apt_ppa.get_launchpad_ppa_key_id",
        spec=apt_ppa.get_launchpad_ppa_key_id,
        return_value="FAKE-PPA-SIGNING-KEY",
    )


@pytest.fixture(autouse=True)
def mock_apt_uca_key_id(mocker):
    yield mocker.patch.object(package_repository, "UCA_KEY_ID", "FAKE-UCA-KEY-ID")


@pytest.fixture()
def mock_logger(mocker):
    yield mocker.patch(
        "craft_archives.repo.apt_key_manager.logger", spec=logging.Logger
    )


@pytest.fixture
def key_assets(tmp_path):
    assets = tmp_path / "key-assets"
    assets.mkdir(parents=True)
    yield assets


@pytest.fixture
def apt_gpg(key_assets, tmp_path):
    yield AptKeyManager(
        keyrings_path=tmp_path,
        key_assets=key_assets,
    )


def test_find_asset(
    apt_gpg,
    key_assets,
):
    key_id = "8" * 40
    expected_key_path = key_assets / ("8" * 8 + ".asc")
    expected_key_path.write_text("key")

    key_path = apt_gpg.find_asset_with_key_id(key_id=key_id)

    assert key_path == expected_key_path


def test_find_asset_none(
    apt_gpg,
):
    key_path = apt_gpg.find_asset_with_key_id(key_id="foo")

    assert key_path is None


def test_get_key_fingerprints(
    apt_gpg,
    mock_run,
):
    mock_run.return_value.stdout = SAMPLE_GPG_SHOW_KEY_OUTPUT

    ids = apt_gpg.get_key_fingerprints(key="8" * 40)

    assert ids == ["FAKE-KEY-ID-FROM-GNUPG"]
    assert mock_run.mock_calls == [
        call(
            [
                "gpg",
                "--batch",
                "--no-default-keyring",
                "--with-colons",
                "--homedir",
                mock.ANY,
                "--import-options",
                "show-only",
                "--import",
            ],
            input=b"8" * 40,
            capture_output=True,
            check=True,
            env={"LANG": "C.UTF-8"},
        )
    ]


@pytest.mark.parametrize(
    "should_raise,expected_is_installed",
    [
        (True, False),
        (False, True),
    ],
)
def test_is_key_installed(
    should_raise, expected_is_installed, apt_gpg, mock_run, tmp_path
):
    if should_raise:
        mock_run.side_effect = subprocess.CalledProcessError(
            returncode=2, cmd=[], output=b""
        )
    else:
        mock_run.returncode = 0

    keyring_path = tmp_path / "craft-FOO.gpg"

    # If the keyring file doesn't exist at all the function should exit early,
    # with no gpg calls
    assert not apt_gpg.is_key_installed(key_id="foo")
    assert mock_run.mock_calls == []

    keyring_path.touch()
    is_installed = apt_gpg.is_key_installed(key_id="foo")

    assert is_installed is expected_is_installed
    assert mock_run.mock_calls == [
        call(
            [
                "gpg",
                "--batch",
                "--no-default-keyring",
                "--with-colons",
                "--keyring",
                f"gnupg-ring:{keyring_path}",
                "--list-keys",
                "foo",
            ],
            input=None,
            capture_output=True,
            check=True,
            env={"LANG": "C.UTF-8"},
        )
    ]


@pytest.mark.parametrize("return_code", [1, 2, 130])
def test_is_key_installed_with_gpg_failure(
    apt_gpg, mock_run, mock_logger, tmp_path, return_code
):
    keyring_file = tmp_path / "craft-FOO.gpg"
    keyring_file.touch()
    mock_run.side_effect = subprocess.CalledProcessError(
        cmd=["gpg"], returncode=return_code, output=b"some error"
    )

    is_installed = apt_gpg.is_key_installed(key_id="foo")

    assert is_installed is False
    mock_logger.warning.assert_called_once_with("gpg error: some error")


def test_install_key(
    apt_gpg, mock_run, mock_chmod, sample_key_string, sample_key_bytes
):
    mock_run.return_value.stdout = SAMPLE_GPG_SHOW_KEY_OUTPUT

    apt_gpg.install_key(key=sample_key_string)

    assert mock_run.mock_calls == [
        call(
            [
                "gpg",
                "--batch",
                "--no-default-keyring",
                "--with-colons",
                "--homedir",
                mock.ANY,
                "--import-options",
                "show-only",
                "--import",
            ],
            input=sample_key_bytes,
            capture_output=True,
            check=True,
            env={"LANG": "C.UTF-8"},
        ),
        call(
            [
                "gpg",
                "--batch",
                "--no-default-keyring",
                "--with-colons",
                "--keyring",
                mock.ANY,
                "--import",
                "-",
            ],
            input=sample_key_bytes,
            capture_output=True,
            check=True,
            env={"LANG": "C.UTF-8"},
        ),
    ]


def test_install_key_missing_dir(
    sample_key_string, mock_run, mock_chmod, tmp_path, key_assets
):
    keyrings_path = tmp_path / "keyrings"
    assert not keyrings_path.exists()

    apt_gpg = AptKeyManager(
        keyrings_path=keyrings_path,
        key_assets=key_assets,
    )
    mock_run.return_value.stdout = SAMPLE_GPG_SHOW_KEY_OUTPUT

    apt_gpg.install_key(key=sample_key_string)
    assert keyrings_path.exists()


def test_install_package_repository_key_missing_dir(
    mock_run, mock_chmod, tmp_path, key_assets, mock_apt_ppa_get_signing_key
):
    keyrings_path = tmp_path / "keyrings"
    assert not keyrings_path.exists()
    repo = PackageRepositoryAptPPA(ppa="snappy-dev/snapcraft-daily", type="apt")
    mock_run.return_value.stdout = SAMPLE_GPG_SHOW_KEY_OUTPUT
    apt_gpg = AptKeyManager(keyrings_path=keyrings_path, key_assets=key_assets)

    apt_gpg.install_package_repository_key(package_repo=repo)

    assert keyrings_path.is_dir()


def test_install_key_with_gpg_failure(apt_gpg, mock_run):
    mock_run.side_effect = [
        subprocess.CompletedProcess(
            ["gpg", "--do-something"], returncode=0, stdout=SAMPLE_GPG_SHOW_KEY_OUTPUT
        ),
        subprocess.CalledProcessError(cmd=["foo"], returncode=1, output=b"some error"),
    ]

    with pytest.raises(errors.AptGPGKeyInstallError) as raised:
        apt_gpg.install_key(key="FAKEKEY")

    assert str(raised.value) == "Failed to install GPG key: some error"


@pytest.mark.parametrize(
    "fingerprints,error",
    [
        pytest.param([], "Invalid GPG key", id="no keys"),
        pytest.param(
            ["finger1", "finger2"],
            "Key must be a single key, not multiple.",
            id="multiple keys",
        ),
    ],
)
def test_install_key_with_key_issue(apt_gpg, mocker, fingerprints, error):
    mock_fingerprints = mocker.patch.object(apt_gpg, "get_key_fingerprints")
    mock_fingerprints.return_value = fingerprints

    with pytest.raises(errors.AptGPGKeyInstallError) as raised:
        apt_gpg.install_key(key="key")

    assert str(raised.value) == f"Failed to install GPG key: {error}"


def test_install_key_from_keyserver(apt_gpg, mock_run, mock_chmod):
    apt_gpg.install_key_from_keyserver(key_id="FAKE_KEY", key_server="key.server")

    assert mock_run.mock_calls == [
        call(
            [
                "gpg",
                "--batch",
                "--no-default-keyring",
                "--with-colons",
                "--keyring",
                mock.ANY,
                "--homedir",
                mock.ANY,
                "--keyserver",
                "key.server",
                "--recv-keys",
                "FAKE_KEY",
            ],
            check=True,
            env={"LANG": "C.UTF-8"},
            input=None,
            capture_output=True,
        )
    ]
    # Two chmod calls: one for the temporary dir that gpg uses during the fetching,
    # and one of the actual keyring file.
    assert mock_chmod.mock_calls == [call(mock.ANY, 0o700), call(mock.ANY, 0o644)]


def test_install_key_from_keyserver_with_apt_key_failure(apt_gpg, mock_run):
    mock_run.side_effect = subprocess.CalledProcessError(
        cmd=["gpg"], returncode=1, output=b"some error"
    )

    with pytest.raises(errors.AptGPGKeyInstallError) as raised:
        apt_gpg.install_key_from_keyserver(
            key_id="fake-key-id", key_server="fake-server"
        )

    assert str(raised.value) == "Failed to install GPG key: some error"


@pytest.mark.parametrize(
    "is_installed",
    [True, False],
)
def test_install_package_repository_key_already_installed(
    is_installed, apt_gpg, mocker, mock_chmod
):
    mocker.patch(
        "craft_archives.repo.apt_key_manager.AptKeyManager.is_key_installed",
        return_value=is_installed,
    )
    package_repo = PackageRepositoryApt(
        type="apt",
        components=["main", "multiverse"],
        key_id="8" * 40,
        key_server="xkeyserver.com",
        suites=["xenial"],
        url="http://archive.ubuntu.com/ubuntu",
    )

    updated = apt_gpg.install_package_repository_key(package_repo=package_repo)

    assert updated is not is_installed


def test_install_package_repository_key_from_asset(apt_gpg, key_assets, mocker):
    mocker.patch(
        "craft_archives.repo.apt_key_manager.AptKeyManager.is_key_installed",
        return_value=False,
    )
    mock_install_key = mocker.patch(
        "craft_archives.repo.apt_key_manager.AptKeyManager.install_key"
    )

    key_id = "123456789012345678901234567890123456AABB"
    expected_key_path = key_assets / "3456AABB.asc"
    expected_key_path.write_text("key-data")

    package_repo = PackageRepositoryApt(
        type="apt",
        components=["main", "multiverse"],
        key_id=key_id,
        suites=["xenial"],
        url="http://archive.ubuntu.com/ubuntu",
    )

    updated = apt_gpg.install_package_repository_key(package_repo=package_repo)

    assert updated is True
    assert mock_install_key.mock_calls == [call(key="key-data")]


def test_install_package_repository_key_apt_from_keyserver(apt_gpg, mocker):
    mock_install_key_from_keyserver = mocker.patch(
        "craft_archives.repo.apt_key_manager.AptKeyManager.install_key_from_keyserver"
    )
    mocker.patch(
        "craft_archives.repo.apt_key_manager.AptKeyManager.is_key_installed",
        return_value=False,
    )

    key_id = "8" * 40

    package_repo = PackageRepositoryApt(
        type="apt",
        components=["main", "multiverse"],
        key_id=key_id,
        key_server="key.server",
        suites=["xenial"],
        url="http://archive.ubuntu.com/ubuntu",
    )

    updated = apt_gpg.install_package_repository_key(package_repo=package_repo)

    assert updated is True
    assert mock_install_key_from_keyserver.mock_calls == [
        call(key_id=key_id, key_server="key.server")
    ]


def test_install_package_repository_key_ppa_from_keyserver(apt_gpg, mocker):
    mock_install_key_from_keyserver = mocker.patch(
        "craft_archives.repo.apt_key_manager.AptKeyManager.install_key_from_keyserver"
    )
    mocker.patch(
        "craft_archives.repo.apt_key_manager.AptKeyManager.is_key_installed",
        return_value=False,
    )

    package_repo = PackageRepositoryAptPPA(type="apt", ppa="test/ppa")
    updated = apt_gpg.install_package_repository_key(package_repo=package_repo)

    assert updated is True
    assert mock_install_key_from_keyserver.mock_calls == [
        call(key_id="FAKE-PPA-SIGNING-KEY", key_server="keyserver.ubuntu.com")
    ]


def test_install_package_repository_key_uca_from_keyserver(apt_gpg, mocker):
    mock_install_key_from_keyserver = mocker.patch(
        "craft_archives.repo.apt_key_manager.AptKeyManager.install_key_from_keyserver"
    )
    mocker.patch(
        "craft_archives.repo.apt_key_manager.AptKeyManager.is_key_installed",
        return_value=False,
    )

    package_repo = PackageRepositoryAptUCA(type="apt", cloud="antelope")
    updated = apt_gpg.install_package_repository_key(package_repo=package_repo)

    assert updated is True
    assert mock_install_key_from_keyserver.mock_calls == [
        call(key_id="FAKE-UCA-KEY-ID", key_server="keyserver.ubuntu.com")
    ]


def test_keyrings_path_for_root():
    assert AptKeyManager.keyrings_path_for_root() == KEYRINGS_PATH
    assert AptKeyManager.keyrings_path_for_root(Path("/my/root")) == Path(
        "/my/root/etc/apt/keyrings"
    )



# This file is part of craft-archives.
#
# Copyright 2023 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Common errors used in craft-archives."""

import dataclasses
from typing import Optional


@dataclasses.dataclass(repr=True)
class ArchivesError(Exception):
    """Unexpected error.

    :param brief: Brief description of error.
    :param details: Detailed information.
    :param resolution: Recommendation, if any.
    """

    brief: str
    details: Optional[str] = None
    resolution: Optional[str] = None

    def __str__(self) -> str:
        return self.brief



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2019-2023 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Project model definitions and helpers."""
from typing import Any, Dict

from craft_archives.repo.package_repository import PackageRepository


def validate_repository(data: Dict[str, Any]) -> None:
    """Validate a package repository.

    :param data: The repository data to validate.
    """
    if not isinstance(data, dict):  # pyright: ignore[reportUnnecessaryIsInstance]
        raise TypeError("value must be a dictionary")
    PackageRepository.unmarshal(data)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2023 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Integration tests for AptKeyManager"""


import pytest
from craft_archives.repo.apt_key_manager import AptKeyManager


@pytest.fixture
def key_assets(tmp_path):
    assets = tmp_path / "key-assets"
    assets.mkdir(parents=True)
    return assets


@pytest.fixture
def gpg_keyring(tmp_path):
    return tmp_path / "keyring.gpg"


@pytest.fixture
def apt_gpg(key_assets, tmp_path):
    return AptKeyManager(
        keyrings_path=tmp_path,
        key_assets=key_assets,
    )


def test_install_key(apt_gpg, tmp_path, test_data_dir):
    expected_file = tmp_path / "craft-FC42E99D.gpg"
    assert not expected_file.exists()
    assert not apt_gpg.is_key_installed(key_id="FC42E99D")

    keypath = test_data_dir / "FC42E99D.asc"
    apt_gpg.install_key(key=keypath.read_text())

    assert expected_file.is_file()
    assert apt_gpg.is_key_installed(key_id="FC42E99D")

    # Check that gpg's backup file has been removed
    backup_file = expected_file.with_suffix(expected_file.suffix + "~")
    assert not backup_file.is_file()


def test_install_key_from_keyserver(apt_gpg, tmp_path):
    expected_file = tmp_path / "craft-FC42E99D.gpg"
    assert not expected_file.exists()
    assert not apt_gpg.is_key_installed(key_id="FC42E99D")

    key_id = "78E1918602959B9C59103100F1831DDAFC42E99D"
    apt_gpg.install_key_from_keyserver(key_id=key_id)

    assert expected_file.is_file()
    assert apt_gpg.is_key_installed(key_id="FC42E99D")


def test_install_key_missing_directory(key_assets, tmp_path, test_data_dir):
    keyrings_path = tmp_path / "keyrings"
    assert not keyrings_path.exists()

    apt_gpg = AptKeyManager(
        keyrings_path=keyrings_path,
        key_assets=key_assets,
    )

    keypath = test_data_dir / "FC42E99D.asc"
    apt_gpg.install_key(key=keypath.read_text())

    assert keyrings_path.exists()
    assert keyrings_path.stat().st_mode == 0o40755  # noqa: PLR2004 magic value
